#!/usr/local/bin/perl 
#!/usr/bin/perl -X For Production

# cvsmonitor.pl - Monitor activity on a CVS Repository
# Copyright (C) 2002-2004 Adam Kennedy
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Should you wish to utilise this software under a different licence, 
# please contact the author.

# Set up the Perl core environment
use vars qw{$DEBUG};
BEGIN {
	require 5.005;

	# Master DEBUG flag
	$DEBUG = 0;
}
use strict;
# use warnings; # Not supported before 5.6
use UNIVERSAL 'isa';

# CVSMonitor does NOT work with mod_perl. To prevent people from
# trying ( THAT MEANS YOU ), die if we detect it. Don't try to make
# it work, if it was easy, it would have been done already.
BEGIN {
	if ( $ENV{MOD_PERL} ) {
		die "CVSMonitor will not work under mod_perl. "
		. "Please execute as a normal CGI script.";
	}
}

# Always run from the directory the script is in
use FindBin; BEGIN { chdir $FindBin::Bin }

# Specify the location of the other libraries
BEGIN {
	@INC = ( 'modules', @INC, 'bundled' );
}

# Load in the modules we will need
use Time::Duration ();
use URI::Escape ();
use File::Spec ();
use CGI ();
use CGI::Carp qw{fatalsToBrowser};
BEGIN {
	# Tweaks to CGI behaviour
	$CGI::DEBUG = 2 if $DEBUG; 
	$CGI::DEBUG; # Avoids a 'only used once' warning. 

	# In the case that we are behind a proxy that will change HTTP_HOST,
	# and we need to generate full links, the old, correct, value for
	# HTTP_HOST _might_ be stored in HTTP_X_FORWARDED_HOST.
	# The only feature that needs this at this time is RSS.
	# We want to use these forwarded values instead of the normal ones.
	$ENV{HTTP_HOST} = $ENV{HTTP_X_FORWARDED_HOST} if defined $ENV{HTTP_X_FORWARDED_HOST};
}

use Class::Autouse; 
BEGIN {
	# Turn on devel mode if debugging
	Class::Autouse->devel(1) if $DEBUG;
}
use Class::Autouse 'File::Flat',
                   'Array::Window',
                   'XML::Generator';
use AppLib ();
use CVSMonitor ();

# Initialise globals and modules
use vars qw{$VERSION $versionstring};                 # Fundamentals
use vars qw{$configfile $Config $static $callback};   # Configuration
use vars qw{$cmdmap};                                 # Data Sets
use vars qw{%in $fin $cmd $message $gui $imagemode};  # Call params
use vars qw{$MetaData $Repository $Module $rep $mod}; # Data Stores
use vars qw{$tab_width $TAB};                         # Config option not integrated yet
use vars qw{$execfile $htmldir};                      # Various paths
BEGIN {
	# Set the static and configuration
	$VERSION = 0.6;
	$versionstring = '0.6.3';
	$configfile = File::Spec->catfile( File::Spec->curdir, 'main.conf' );
	$callback = $ENV{SCRIPT_NAME};

	$tab_width = 8;
	$TAB = ' ' x $tab_width;

	$execfile = File::Spec->catfile( File::Spec->curdir, 'cvsexec.pl' );
	$htmldir  = File::Spec->catfile( File::Spec->curdir, 'html' );
}

# Main application logic
initialise();
doCall();
exit();

# Application initialisation
sub initialise {
	$cmdmap = {
		# Normal pages
		viewAbout                       => \&viewAbout,

		# Configuration
		viewConfigMissing               => \&viewConfigMissing,
		viewConfigEdit                  => \&viewConfigEdit,
		configSave                      => \&cmdConfigSave,

		# Admin Section
		viewAdmin                       => \&viewAdmin,
		viewAdminRepositoryList         => \&viewAdminRepositoryList,
		viewAdminModuleDetails          => \&viewAdminModuleDetails,
		viewAdminModuleNew              => \&viewAdminModuleNew,
		viewAdminRepositoryNew          => \&viewAdminRepositoryNew,
		viewAdminModuleErrorLog         => \&viewAdminModuleErrorLog,
		viewAdminLogList                => \&viewAdminLogList,
		viewAdminLog                    => \&viewAdminLog,
		adminNewModule                  => \&cmdAdminNewModule,
		adminDeleteModule               => \&cmdAdminDeleteModule,
		adminNewRepositoryTest          => \&cmdAdminNewRepositoryTest,
		adminNewRepository              => \&cmdAdminNewRepository,
		adminDeleteRepository           => \&cmdAdminDeleteRepository,
		adminNewTracker                 => \&cmdAdminNewTracker,
		adminDeleteTracker              => \&cmdAdminDeleteTracker,
		adminRelease                    => \&cmdAdminRelease,
		adminRecover                    => \&cmdAdminRecover,
		adminBackgroundCheckout         => \&cmdAdminBackgroundCheckout,
		adminBackgroundUpdate           => \&cmdAdminBackgroundUpdate,
		adminBackgroundUpdateRepository => \&cmdAdminBackgroundUpdateRepository,
		adminBackgroundUpdateAll        => \&cmdAdminBackgroundUpdateAll,
		adminResetMetaDataCache         => \&cmdAdminResetMetaDataCache,
		adminClearLogs                  => \&cmdAdminClearLogs,

		# Browse section
		viewBrowse                      => \&viewBrowse,
		viewBrowseRepository            => \&viewBrowseRepository,
		viewBrowseModule                => \&viewBrowseModule,
		viewBrowseYear                  => \&viewBrowseYear,
		viewBrowseMonth                 => \&viewBrowseMonth,
		viewBrowseDay                   => \&viewBrowseDay,
		viewBrowseChangeSet             => \&viewBrowseChangeSet,
		viewBrowseFile                  => \&viewBrowseFile,
		viewBrowseFileHistory           => \&viewBrowseFileHistory,
		viewBrowseFileAnnotation        => \&viewBrowseFileAnnotation,
		viewBrowseFileBlame             => \&viewBrowseFileBlame,
		# viewBrowseFileBranches        => \&viewBrowseFileBranches,
		viewBrowseVersion               => \&viewBrowseVersion,
		viewBrowseVersionDiff           => \&viewBrowseVersionDiff,
		viewBrowseAuthor                => \&viewBrowseAuthor,

		# Graph Related
		viewGrapher                     => \&viewGrapher,
		graphModuleLines                => \&graphModuleLines,
		graphModuleKarma                => \&graphModuleKarma,
		graphModuleFiles                => \&graphModuleFiles,
		graphRepositoryLines            => \&graphRepositoryLines,
		graphRepositoryKarma            => \&graphRepositoryKarma,
		graphRepositoryFiles            => \&graphRepositoryFiles,
		graphAuthorLines                => \&graphAuthorLines,
		graphAuthorKarma                => \&graphAuthorKarma,
		graphAuthorFiles                => \&graphAuthorFiles,

		# Search
		viewSearch                      => \&viewSearch,
		viewSearchChangeSet             => \&viewSearchChangeSet,
		viewSearchFile                  => \&viewSearchFile,
		searchChangeSet                 => \&cmdSearchChangeSet,
		searchFile                      => \&cmdSearchFile,

		# RSS
		rssModule                       => \&rssModule,
		};

	# Make sure cvsexec exists
	unless ( -f $execfile ) {
		Error( "The required file cvsexec.pl appears to be missing from your directory" );
	}

	# Make sure cvsexec is executable
	unless ( -x $execfile ) {
		# Can we make it executable
		chmod( $execfile, 0755);
		unless ( -x $execfile ) {
			Error( "The script cvsexec.pl needs to be executable by the web user in order"
				. " for this application to function correctly.<br>Please correct this"
				. " and try again" );
		}
	}

	# Initialise the page library
	AppLib::PageFactory->setBasePath( $htmldir );

	# Try to load the Config object
	$Config = CVSMonitor::Config->load( $configfile );
	if ( $Config ) {
		CVSMonitor::Security->setPassword( $Config->getSecurityAdminpassword );
	}

	# Don't initialise the gui if we don't have a config
	initialise_gui() if $Config;
}

sub initialise_gui {
	return unless $Config;

	# Set the static location information
	$static = {
		filesystem => $Config->getHostStaticfilesystem,
		http       => $Config->getHostStatichttp,
		};

	# Now we know where the static stuff is, add the page filter
	# to all generated pages
	AppLib::PageFactory->addFilter( 
		sub {
			my $Page = shift;
			my $content = shift;

			# Find all the references to ../static, and change to
			# the static page path
			$content =~ s!\.\.[\.\/]+static/!$static->{http}/static/!g;

			$content;
		} );
}

# Process the call. ( Like doGet/goPost in Java )
sub doCall {
	# Get the CGI data
	CGI::ReadParse();
	$fin = { %in };

	# If we don't have a config file, don't do anything that isn't related to it
	unless ( $Config ) {
		unless ( defined $fin->{cmd} and $fin->{cmd} =~ /^(?:viewConfig|config)/ ) {
			# Go to the "You don't have a config" page.
			$fin->{cmd} = 'viewConfigMissing';
		}
	}

	# Apply the default call
	$cmd = $fin->{cmd} ? $fin->{cmd} : 'viewBrowse';

	# If this is a graph call, split off now
	if ( $cmd =~ /^graph/ ) {
		return doGraph();
	}

	# Set up the gui data tree, and prepare everything.
	# Using this simple hash is a precursor to later doing this
	# as a session object or something.
	$gui = {
		title    => '',
		toolbars => {},
		};

	# Unless we are doing something funny, load the MetaData cache
	unless ( $cmd =~ /^(?:viewConfig|config|adminResetMetaDataCache)/ ) {
		checkMetaData();
	}

	# Check that the command exists
	unless ( defined $cmdmap->{$cmd} ) {
		addToolbar( 'main' );
		Error( "Invalid script action '$cmd'" );
	}

	# Handle RSS mode
	return &{ $cmdmap->{$cmd} } if $cmd =~ /^rss/;

	# Add the main toolbar in normal mode
	addToolbar( 'main' );

	my $result = undef;
	if ( ($cmd =~ /^admin/ or $cmd =~ /^viewAdmin/) and ! authenticate() ) {
		# The script needed authentication and failed
		$result = viewAdminLogin();

	} else {
		# If this is an Admin action, make sure the MetaData cache is not active
		if ( $cmd =~ /^admin/ ) {
			$result = doAdminTask();
		} else {
			# Call the function
			$result = &{ $cmdmap->{$cmd} };
		}
	}
	Error( "$cmd returned a negative value" ) unless $result;

	# Handle the result of the call
	if ( isa( $result, 'AppLib::Page' ) ) {
		# Show the resulting page
		unless ( $result->display ) {
			Error( "Error trying to show page<br>" . AppLib::Error->errstrHTML );
		}
		return 1;

	} elsif ( isa( $result, 'CVSMonitor::Graph' ) ) {
		# Send the Graph
		$result->sendToBrowser;
		return 1;

	} else {
		# If they otherwise return true, do nothing
	}
}

# Special call handler for graphs
sub doGraph {
	# We don't need to deal with pages or guis here, because all we
	# do is spit back an image. We also don't need any security, 
	# because there's no way they can get a page.
	$imagemode = 1;

	# Check the MetaData
	checkMetaData();

	# Does the command exist
	unless ( defined $cmdmap->{$cmd} ) {
		Error( "Invalid script action '$cmd'" );
	}

	# Call the graph generator
	my $result = &{ $cmdmap->{$cmd} };
	ErrorImage() unless isa( $result, 'CVSMonitor::Graph' );

	# Display the image
	$result->display or ErrorImage();

	1;
}

sub doAdminTask {
	# Before we can do any admin task, we have to make sure that the MetaData
	# is not active. That is, nobody else is doing anything.
	if ( 
		$cmd !~ /^adminNew/
		and defined $MetaData 
		and $MetaData->active 
	) {
		$message = 'Another task is currently modifying the MetaData cache<br>'
			. 'Please wait until it has completed, and try again';
		return viewAdmin();
	}

	# Look ok, now do the task as normal
	&{ $cmdmap->{$cmd} };
}

sub authenticate {
	# Did they give us an explicit password
	if ( $fin->{_admin_pass} ) {
		my $rv = CVSMonitor::Security->authenticate( $fin->{_admin_pass} );
		if ( $rv ) {
			# Save the password
			CVSMonitor::Security->savePassword( $fin->{_admin_pass} );
			return 1;
		} elsif ( defined $rv ) {
			# No password
			$message = "You did not enter a password";
		} else {
			# Bad password
			$message = "Incorrect password";
		}
	} else {
		my $rv = CVSMonitor::Security->authenticate();
		return 1 if $rv;
		unless ( defined $rv ) {
			# Password in cookie is different
			$message = "The password does not match the password you have stored.<br>"
				. "Please enter the password again";
		}
	}

	0;
}





#####################################################################
# Admin Actions

# Add a new Module underneath a Repository
sub cmdAdminNewModule {
	# Check and find the repository name
	unless ( $fin->{repository} ) {
		$message = 'You did not select a repository';
		return viewAdminModuleNew();
	}

	my $Repository = $MetaData->getRepository( $fin->{repository} );
	unless ( $Repository ) {
		$message = "The Repository '$fin->{repository}' does not exist";
		return viewAdminModuleNew();
	}

	# Get and check the name
	unless ( $fin->{name} ) {
		$message = "You did not enter a symbolic name for the module";
		return viewAdminModuleNew();
	}
	unless ( $MetaData->_checkFormat( 'symbol', $fin->{name} ) ) {
		$message = "The symbolic name is invalid<br>It should consist of 1-21 alphanumeric characters";
		return viewAdminModuleNew();
	}

	# Do we already have a module by that name
	if ( $MetaData->getModule($fin->{module}) ) {
		$message = "A module by that name has already been added";
		return viewAdminModuleNew();
	}

	# Check the module path
	unless ( $fin->{path} ) {
		$message = "You did not enter the path of the module in the repository";
		return viewAdminModuleNew();
	}

	# Try to add the module
	unless ( $Repository->addModule( $fin->{name}, $fin->{path} ) ) {
		Error( $MetaData->errstr );
	}

	# Save the index
	$MetaData->save or Error( "Failed to save changes" );

	# We added the module, show it's page
	$message = "Module '$fin->{name}' Added. You probably want to "
		. fragmentLink( "CheckOut", {
			cmd    => 'adminBackgroundCheckout',
			module => $Repository->name . ".$fin->{name}",
			} )
		. " the Module now.";

	viewAdmin();
}

# Attempt to add a new Repository
sub cmdAdminNewRepository {
	my $cvsroot = $fin->{cvsroot};

	# Try to create and add the repository, catching errors
	unless ( $MetaData->addRepository( $fin->{name}, $cvsroot, {
		label => $fin->{label},
		details => $fin->{details},
		} ) ) {
		$message = CVSMonitor::MetaData->errstr 
			|| 'Unknown error while trying to create repository';
		return viewAdminRepositoryNewDetails();
	}

	# Save the index
	$MetaData->save or Error( "Failed to save changes" );

	# Done
	$message = "Repository '$fin->{label}' added successfully";
	viewAdmin();
}

# Add a module tracker
sub cmdAdminNewTracker {
	checkModule();

	# Check for a keyword and url
	unless ( $fin->{tracker_keyword} ) {
		$message = 'You did not enter a keyword';
		return viewAdminModuleDetails();
	}
	unless ( $fin->{tracker_url} ) {
		$message = 'You did not enter a url';
		return viewAdminModuleDetails();
	}

	# Try to add the tracker
	unless ( $Module->addTracker( $fin->{tracker_keyword}, $fin->{tracker_url} ) ) {
		$message = 'Failed to add Tracker: ' . $Module->errstr;
		return viewAdminModuleDetails();
	}

	# Tracker added, save and show the details page
	$MetaData->save or Error( 'Failed to save changes to MetaData' );

	$message = "Tracker '$fin->{tracker_keyword}' added";
	delete $fin->{tracker_keyword};
	delete $fin->{tracker_url};
	viewAdminModuleDetails();
}

# Remove a module tracker
sub cmdAdminDeleteTracker {
	checkModule();

	# Check for keyword
	unless ( $fin->{keyword} ) {
		$message = 'You did not provide a keyword to delete';
		return viewAdminModuleDetails();
	}

	# Delete the keyword
	$Module->deleteTracker( $fin->{keyword} );

	# Save and return
	$MetaData->save or Error( 'Failed to save changes to MetaData' );

	$message = "Tracker '$fin->{keyword}' deleted";
	viewAdminModuleDetails();
}

sub cmdAdminBackgroundCheckout {
	checkModule();

	# Release the lock on the index file
	$MetaData->releaseLock;

	# Detachably call the exec script to do the checkout
	system( "$FindBin::Bin/cvsexec.pl -v -l checkout $mod 1>/dev/null &" );

	# Wait for a few seconds to let it fire up...
	sleep(3);

	# Reload the index, but don't take a lock
	$MetaData->reload( 'nolock' )
		or Error( "Appeared to background OK, but failed to reload the index" );

	# Set the message and show the admin view
	$message = "Backgrounded checkout started for Module '$mod'";
	viewAdmin();
}

sub cmdAdminBackgroundUpdate {
	checkModule();

	# Release the lock on the index file
	$MetaData->releaseLock;

	# Detachably call the exec script to do the update
	system( "$FindBin::Bin/cvsexec.pl -v -l update $mod 1>/dev/null &" );

	# Wait for a few seconds to let it fire up...
	sleep(3);

	# Reload the index, but don't take a lock
	$MetaData->reload( 'nolock' )
		or Error( "Appeared to background OK, but failed to reload the index" );

	# Set the message and show the admin view
	$message = "Backgrounded update started for Module '$mod'";
	viewAdmin();
}

sub cmdAdminBackgroundUpdateRepository {
	checkRepository();

	# Release the lock on the index file
	$MetaData->releaseLock;

	# Detachably call the exec script to do the update
	system( "$FindBin::Bin/cvsexec.pl -v -l updaterepository $rep 1>/dev/null &" );

	# Wait for a few seconds to let it fire up...
	sleep(3);

	# Reload the index, but don't take a lock
	$MetaData->reload( 'nolock' )
		or Error( "Appeared to background OK, but failed to reload the index" );

	# Set the message and show the admin view
	$message = "Backgrounded update started for Repository '$rep'";
	viewAdmin();
}

sub cmdAdminBackgroundUpdateAll {
	# Release the lock on the index file
	$MetaData->releaseLock;

	# Detachably call the exec script to do the update all
	system( "$FindBin::Bin/cvsexec.pl -v -l updateall 1>/dev/null &" );

	# Wait for a few seconds to let it fire up...
	sleep(3);

	# Reload the index, but don't take a lock
	$MetaData->reload( 'nolock' )
		or Error( "Appeared to background OK, but failed to reload the index" );

	# Set the message and show the admin view
	$message = "Backgrounded update started for all Modules.";
	viewAdmin();
}

sub cmdAdminRelease {
	checkModule();

	# Release the module
	unless ( $Module->release ) {
		Error( "Error while releasing: " . $MetaData->errstr );
	}

	# Save the index and reload the view
	$MetaData->save;
	$message = "Module '$mod' Released";
	viewAdmin();
}

# Recover a Module from it's broken state
sub cmdAdminRecover {
	checkModule();

	# Take control of the module
	unless ( $Module->lock ) {
		Error( "Error trying to take Module lock: " . $MetaData->errstr );
	}

	# Purge the file system
	unless ( $Module->purge ) {
		Error( "Error trying to purge disk for Module: " . $MetaData->errstr );
	}

	# Reset it's state, release the module, and save the index
	$Module->set( 'state', 'Unavailable' );
	$Module->set( 'date', undef );
	$Module->set( 'updateTime', undef );
	$Module->set( 'footprint', undef );
	$Module->releaseLock or Error( "Error releasing Module" );
	$MetaData->save or Error( "Error saving Index" );

	# Set the message and go to the admin view
	my $display = $Module->display;
	$message = "Recovered Module '$display'";
	viewAdmin();
}

sub cmdAdminDeleteModule {
	# If the MetaData is already active, go back to the RepositoryList
	# page and tell them to wait.
	if ( $MetaData->active ) {
		$message = 'Another task is currently modifying the MetaData cache<br>'
			. 'Please wait until it has completed, and try again';
		return viewAdminRepositoryList();
	}

	# Is the module valid
	checkModule();

	# Delete the module from it's repository
	unless ( $Module->delete ) {
		Error( "Error while deleting: " . $MetaData->errstr );
	}

	# Save the index and reload the view
	$MetaData->save or Error( "Error saving Index" );
	$message = "Module '$mod' Deleted";
	viewAdmin();
}

sub cmdAdminDeleteRepository {
	checkRepository();

	# Get some values for later
	my $label = $Repository->getLabel;

	# Try to delete the repository
	unless ( $MetaData->deleteRepository( $rep ) ) {
		Error( "Failed to delete Repository.<br>" . CVSMonitor::MetaData->errstr );
	}

	# Save the index and reload the view
	$MetaData->save or Error( "Error saving index after repository deletion" );
	$message = "Repository '$label' deleted";
	viewAdmin();
}

# Attempt to reset the cache, the method will cache all the 
# important errors.
sub cmdAdminResetMetaDataCache {
	# Release the old cache if applicable
	if ( $MetaData ) {
		$MetaData->releaseLock;
		undef $MetaData;
	}

	# Remove the old MetaData cache directory
	my $rv = CVSMonitor::MetaData->removeMetaDataCache( $Config->getHostScratch );
	unless ( $rv ) {
		Error( "Error while trying to remove the old MetaData cache" );
	}

	# Create a new MetaData cache
	$MetaData = CVSMonitor::MetaData->new( $Config->getHostScratch );
	unless ( $MetaData ) {
		Error( "Error while creating new, replacement, MetaData cache" );
	}

	# Done
	$message = "Removed and recreated MetaData cache";
	viewAdmin();
}

sub cmdAdminClearLogs {
	# Remove the log directory
	if ( File::Flat->remove( $MetaData->pathAdminLogDir ) ) {
		$message = 'Log directory cleared';
		return viewAdminRepositoryList();
	} else {
		Error( "Failed to delete log directory: " . File::Flat->errstr );
	}
}

sub cmdConfigSave {
	# To do this, either they have to be authenticated, or
	# there is no Config file
	if ( $Config ) {
		return viewAdminLogin() unless authenticate();
	} else {
		# Set the Config to a new Config
		$Config = CVSMonitor::Config->new() 
			or Error( "Failed to create new config" );
	}

	# Create the issues list
	my @issues = ();

	# Set the time format
	unless ( $Config->setGuiTimeformat( $fin->{gui_timeformat} ) ) {
		push @issues, "Failed to set time format";
	}

	# Set the ChangeSet quantity
	$Config->setGuiChangesets( $fin->{gui_changesets} );

	# Set the cache location
	unless ( $Config->setHostStaticfilesystem( $fin->{gui_cachefilesystem} ) ) {
		push @issues, $Config->errstr;
	}
	unless ( $Config->setHostStatichttp( $fin->{gui_cachehttp} ) ) {
		push @issues, $Config->errstr;
	}

	# Set the password
	unless ( $Config->setSecurityAdminpassword( $fin->{admin_password} ) ) {
		push @issues, "Failed to set admin password";
	}

	# Set the working directory
	unless ( $Config->setHostScratch( $fin->{metadata_directory} ) ) {
		push @issues, "Failed to set working directory: " . $Config->errstr;
	}

	# Feed back to the edit page if there were any issues
	if ( @issues ) {
		unshift @issues, "There were problems with your settings";
		$message = join '<br>', @issues;
		return viewConfigEdit();
	}

	# Looks good, save and show the "changed" page
	$Config->save( $configfile ) or Error( "Error saving config", $Config->errstr );

	# Re-initialise the MetaData
	checkMetaData();

	# Re-initialise the gui
	initialise_gui();

	# Add the main toolbar again
	addToolbar( 'main' );

	# Show the admin page
	$message = "Configuration saved to '$configfile'";
	viewAdmin();
}

# Search for a ChangeSet
sub cmdSearchChangeSet {
	# Find the search space
	my $Activity = findSearchSpace();
	unless ( $Activity ) {
		Error( "Could not determine or lost the search space" );
	}

	# If they specified a date, filter by the date
	if ( $fin->{search_date} ) {
		$Activity->filter( 'date', $fin->{search_date} )
			or Error( 'Error while filtering the date' );
	}

	# If they specified an author, filter by the author
	if ( $fin->{search_author} ) {
		$Activity->filter( 'author', $fin->{search_author} );
	}

	# If they specified a commit message search, filter by it
	if ( $fin->{search_message} ) {
		if ( $fin->{search_message_type} eq 'regex' ) {
			unless ( 
				$Activity->filter( 'message', $fin->{search_message}, { 
					type => 'regex', 
					case => $fin->{search_message_case},
					} ) 
			) {
				Error( 'Error while applying regex filter. Your regex may be invalid' );
			}
		} else {
			$Activity->filter( 'message', $fin->{search_message}, {
				type => 'literal',
				case => $fin->{search_message_case}
				} );
		}
	}

	# Act depending on the number of results
	if ( $Activity->elements == 0 ) {
		$message = 'There were no ChangeSets found that matched your request';
		return viewSearchChangeSet();
	}

	# Sort the Activity
	$fin->{search_sortby} = 'karma' unless $fin->{search_sortby};
	if ( $fin->{search_sortby} eq 'newest' ) {
		$Activity->sort( 'date', 0 );

	} elsif ( $fin->{search_sortby} eq 'oldest' ) {
		$Activity->sort( 'date', 1 );

	} else {
		# Default sort
		$Activity->sort( 'karma' );
	}

	# Show the results page for the range they want
	viewSearchChangeSetResults( $Activity );
}

# Search for a ChangeSet
sub cmdSearchFile {
	# Find the search space
	my $Activity = findSearchSpace();
	unless ( $Activity ) {
		Error( "Could not determine or lost the search space" );
	}

	# If they specified an author, filter by the author
	if ( $fin->{search_author} ) {
		$Activity->filter( 'author', $fin->{search_author} );
	}

	# Now let's get the Files for the Activity
	my $Files = $Activity->getFiles
		or Error( 'Could not find Files in Search Space' );

	# Filter based on filename or path
	if ( $fin->{search_filename} ) {
		# Compile the search regex from the different types
		my $compiled;
		if ( $fin->{search_filename_type} eq 'exact' ) {
			# Escape and force to be just the file name
			my $searchstring = quotemeta( $fin->{search_filename} );
			$searchstring = '\/' . $searchstring . '$';
			$compiled = $fin->{search_filename_case}
				? qr/$searchstring/ : qr/$searchstring/i;

		} elsif ( $fin->{search_filename_type} eq 'file.substring' ) {
			# Escape and make anywhere after the last slash
			my $searchstring = quotemeta( $fin->{search_filename} );
			$searchstring = '\/[^\/]*' . $searchstring . '[^\/]*$';
			$compiled = $fin->{search_filename_case}
				? qr/$searchstring/ : qr/$searchstring/i;

		} elsif ( $fin->{search_filename_type} eq 'path.substring' ) {
			# Escape without addition
			my $searchstring = quotemeta( $fin->{search_filename} );
			$compiled = $fin->{search_filename_case}
				? qr/$searchstring/ : qr/$searchstring/i;

		} elsif ( $fin->{search_filename_type} eq 'regex' ) {
			# Test the regex
			eval qq~
				$compiled = $fin->{search_filename_case}
					? qr/$fin->{search_filename}/
					: qr/$fin->{search_filename}/e
				~;
			return undef if $@;

		} else {
			Error( "Unknown filename search type '$fin->{search_filename_type}'" );
		}

		# Apply the filter
		@$Files = grep { $_->filename =~ $compiled } @$Files;
	}

	# Apply the 'exist in current trunk rule'
	if ( $fin->{search_exists} ) {
		@$Files = grep {
			$_->getRecentHeadVersion->state eq 'Exp'
			} @$Files;
	}

	# Act depending on the number of results
	unless ( scalar @$Files ) {
		$message = 'There were no Files found that matched your request';
		return viewSearchFile();
	}

	# Sort by the appropriate method
	if ( $fin->{search_sortby} eq 'path' ) {
		# Sort by the path
		@$Files = map { $_->[0] }
			sort { $a->[1] cmp $b->[1] or $a->[2] cmp $b->[2] }
			map { [ $_, $_->Module->name, $_->filename] }
			@$Files;

	} elsif ( $fin->{search_sortby} eq 'filename' ) {
		# Schwartzian sort by just the file name
		@$Files = map { $_->[0] }
			sort { $a->[1] cmp $b->[1] or $a->[2] cmp $b->[2] }
			map { [ $_, $_->Module->name, ( $_->filename =~ m!/([^/]+)$! ? $1 : $_->filename ) ] } 
			@$Files;

	} else {
		Error( "'$fin->{search_sortby}' sorting is not yet supported" );
	}

	# Show the results page for the range they want
	viewSearchFileResults( $Files );
}

# Test a CVSROOT for a new repository
sub cmdAdminNewRepositoryTest {
	my $cvsroot = $fin->{cvsroot};

	# Did they enter anything
	unless ( $cvsroot ) {
		$message = 'You did not enter a CVSROOT';
		return viewAdminRepositoryNew();
	}

	# Try to create the repository object
	my $Repository = CVSMonitor::MetaData::Repository->new( 'temprepository', $cvsroot );
	unless ( $Repository ) {
		$message = "There is a problem with the Repository<br>"
			. CVSMonitor::MetaData->errstr;
		return viewAdminRepositoryNew();
	}

	# Check the host
	unless ( $Repository->exists ) {
		$message = "There is a problem with the host/port you specified<br>"
			. CVSMonitor::MetaData->errstr;
		return viewAdminRepositoryNew();
	}

	# Looks good, show the page to let them add the other information
	viewAdminRepositoryNewDetails( $Repository->publicCvsroot );
}





#####################################################################
# General, Home and Help Views

sub viewAbout {
	$gui->{title} = 'About';
	page( 'About' );
}





#####################################################################
# Admin Section Views

# Default admin view if the repository listing
sub viewAdmin {
	viewAdminRepositoryList( @_ );
}

# Tell them to login first, and pass the request through
sub viewAdminLogin {
	my $parser = new_parser();

	# Build a collection of hiddens
	$parser->{hidden} = fragmentHiddenFields();

	# Show the page
	page( 'Admin/Login', $parser );
}

# List the repositories
sub viewAdminRepositoryList {
	my $parser = new_parser();

	# Add the repositories
	$parser->{repository_list} = fragmentAdminRepositoryList();

	# If there are no repositories, put a message here for first timers
	unless ( $parser->{repository_list} ) {
		$parser->{repository_list} = qq~
			<table border=0 cellpadding=10><tr><td>
			You do not have any repositories defined.<br>
			<br>
			First time users may find the <a href="http://ali.as/devel/cvsmonitor/help/quickstart.html" target="_new">QuickStart</a> guide of use.<br>
			<br>
			Otherwise, click on the "Add a Repository" link to add a repository.
			</td></tr></table>
			~;
	}

	# Advise the page if there are any ( non-broken ) running transients
	if ( $MetaData->active ) {
		$parser->{running_transients} = "yes";
		$gui->{title} = 'Refreshing...';
	} else {
		$gui->{title} = 'Repositories and Modules';
	}

	# Show the Refresh Link
	$parser->{admin_right_title} = fragmentAdminRefresh();

	# Show the page
	addAdminMenu( $parser );
	$parser->{admin_section_title} = 'Repositories and Modules';
	page( 'Admin/RepositoryList', $parser );
}

sub viewAdminModuleDetails {
	checkModule();

	# Get the parser
	my $parser = new_parser();

	# Add the various bits of information for the module
	$parser->{module_name} = $Module->name;
	$parser->{module_display} = $Module->display;
	$parser->{module_path} = $Module->get( 'module' );
	$parser->{module_state} = $Module->state;
	$parser->{module_repository_name} = $Module->Repository->getLabel;
	$parser->{module_repository_cvsroot} = $Module->Repository->publicCvsroot;
	$parser->{module_repository_host} = $Module->Repository->host;

	# Add the footprint information
	my $footprint = $Module->footprint;
	$parser->{module_footprint_pretty} = $footprint 
		? AppLib::String->humanFileSize( $footprint )
		: 'Unknown';
	$parser->{module_footprint} = $footprint 
		? commify( $footprint )
		: 'Unknown';

	# Add the update costs
	$parser->{module_update_time} = defined $Module->updateTime
		? Time::Duration::duration( $Module->updateTime, 2 )
		: 'Unknown';
	$parser->{module_update_traffic} = $Module->updateCost
		? AppLib::String->humanFileSize( $Module->updateCost )
		: 'Unkown';

	# Create the list of trackers
	my $trackers = $Module->trackers;
	if ( $trackers ) {
		my @data = ( [ 'Keyword', 'URL', 'Commands' ] );
		foreach ( keys %$trackers ) {
			my $url = CVSMonitor::HTML->escape( $trackers->{$_} );
			$url =~ s!(\%n)!<font color='#000099'><b>$1</b></font>!g;
			push @data, [ $_,
				$url,
				fragmentLink( 'Delete', {
					cmd => 'adminDeleteTracker',
					module => $Module->name,
					keyword => $_,
					} )
				];
		}
		$parser->{module_trackers} = CVSMonitor::HTML::Table->table( \@data, {
			top         => 1,
			topclass    => 'table_header',
			outerwidth  => 0,
			} );
	} else {
		$parser->{module_trackers} = 'No links to Tracker systems';
	}
	$parser->{tracker_keyword} = $fin->{tracker_keyword};
	$parser->{tracker_url} = CVSMonitor::HTML->escape( $fin->{tracker_url} );

	# Show the page
	addAdminMenu( $parser );
	$parser->{admin_section_title} = 'Module Details';
	page( 'Admin/ModuleDetails', $parser );
}

sub viewAdminLogList {
	# Find all the log files in the admin log directory
	my $logdir = $MetaData->pathAdminLogDir;
	my @files = ();
	if ( -e $logdir ) {
		unless ( -r $logdir ) {
			Error( "Insufficient permissions to read log directory '$logdir'" );
		}

		# Get the log files
		opendir( LOGDIR, $logdir )
			or Error( "Failed to open log directory '$logdir' for reading" );
		@files = readdir LOGDIR;
		closedir LOGDIR;

		# Filter down to log files
		@files = grep { /^\d+\.log$/ } @files;
	}

	my $parser = new_parser();
	if ( scalar @files ) {
	        # Expand the information on the files and sort
        	@files = map { [ $_, File::Spec->catfile( $logdir, $_ ) ] } @files;
     			foreach ( @files ) {
	                my @st = stat $_->[1];

 	               # Modify time in seconds past the epoch
        	        $_->[2] = $st[9];

 	               # Raw PID
        	        $_->[3] = $_->[0];
                	$_->[3] =~ s/\D+//g;

 	               # File size in bytes
        	        $_->[4] = $st[7];
        	}
		@files = sort { $b->[2] <=> $a->[2] } @files;

		# Get the first line for each file
		foreach ( @files ) {
			open ( FILE, $_->[1] ) or next;
			my $first = <FILE>;
			close FILE;

			chomp $first;
			$_->[5] = $first;
			if ( $_->[5] =~ /Command - (.+)$/ ) {
				$_->[5] = $1;
			} else {
				$_->[5] = "Unknown Command: $_->[5]";
			}
		}

		# Add the found message
		$message = 'Found ' . scalar(@files) . ' log files';

		# Generate a table showing all the logs
		my @data = ( [ 'Log File', 'Created', 'File Size', 'Command' ] );
		foreach ( @files ) {
			push @data, [
				fragmentLink( $_->[0], {
					cmd => 'viewAdminLog',
					pid => $_->[3],
					} ),
				scalar localtime($_->[2]),
				AppLib::String->humanFileSize( $_->[4] ),
				CVSMonitor::HTML->escape( $_->[5] ),
				];
		}

	        $parser->{log_list} = CVSMonitor::HTML::Table->table( \@data, {
        	        top         => 1,
        	        topclass    => 'table_header',
        	        outerwidth  => 0,
               		} );

	} else {
		$message = 'No log files were found';
	}

	# Show the Refresh Link
	$parser->{admin_right_title} = fragmentAdminRefresh();

	# Show the page
	addAdminMenu( $parser );
	$parser->{admin_section_title} = 'Log Viewer';
	page( 'Admin/LogList', $parser );
}

sub viewAdminModuleErrorLog {
	checkModule();

	# Is it broken?
	unless ( $Module->locked and ! $Module->lockHolderExists ) {
		Error( 'The Module does not appear to be broken' );
	}

	# Get the lock PID
	my $pid = $Module->lockPid or Error( 'Failed to get module lock PID' );

	# Hand off to the log viewer
	viewAdminLog( $pid );
}

sub viewAdminLog {
	# Get the log file
	my $pid = shift || $fin->{pid} or Error( 'Unable to determine a PID' );
	my $logfile = $MetaData->pathAdminLogFile( $pid );

	# Does the file exist?
	unless ( File::Flat->exists( $logfile ) ) {
		Error( "Tried to read error log $logfile, but it doesn't exist" );
	}

	# Can we read it
	unless ( File::Flat->canRead( $logfile ) ) {
		Error( "Tried to read error log $logfile, but I don't have enough permissions to read it" );
	}

	# Read it in
	my $contents = File::Flat->slurp( $logfile );
	Error( "Error while reading log file $logfile" ) unless defined $contents;

	# Escape the contents and add to the parser
	my $parser = new_parser();
	$parser->{log} = CVSMonitor::HTML->escape_multiline( $$contents );

	# Show the page
	addAdminMenu( $parser );
	$parser->{admin_section_title} = 'Log Viewer';
	page( 'Admin/Log', $parser );
}

# Page to add a new Module
# This uses the new Repository structure
sub viewAdminModuleNew {
	my $parser = new_parser();
	addAdminMenu( $parser );
	$parser->{admin_section_title} = 'Add a Module';

	# If there are no repositories, get them to add one first
	unless ( $MetaData->getRepositories ) {
		$message = 'Please add a Repository first';
		return viewAdminRepositoryNew();
	}

	my $repository = shift || $fin->{repository};
	if ( $repository ) {
		# Make sure the repository exists
		unless ( $MetaData->getRepository( $repository ) ) {
			Error( 'Tried to create new module in non-existant repository' );
		}
	}

	# Get the list of repository names and labels
	my @data = map { [ $_->name, $_->getLabel ] } 
		$MetaData->getRepositories;

	# Generate and set the html for the dropbox
	$parser->{repository} = CVSMonitor::HTML::Form->dropbox( 'repository', \@data, $repository );

	# Add some values for the forms
	$parser->{name} = $fin->{name} if exists $fin->{name};
	$parser->{path} = $fin->{path} if exists $fin->{path};

	# Show the page
	page( 'Admin/ModuleNew', $parser );
}

# Page to add a new repository
sub viewAdminRepositoryNew {
	my $parser = new_parser();
	addAdminMenu( $parser );
	$parser->{admin_section_title} = 'Add a Repository';

	# Set the form values
	if ( $fin->{cvsroot} ) {
		# Set to their previous values
		$parser->{$_} = $fin->{$_} 
			foreach qw{cvsroot host port path username password};
	} else {
		# Set the defaults
		$parser->{cvsroot} = ':pserver:username:password@host:/var/cvs';
		$parser->{host} = 'host';
		$parser->{port} = '';
		$parser->{path} = '/var/cvs';
		$parser->{username} = 'username';
		$parser->{password} = 'password';
	}

	# Show the page
	page( 'Admin/RepositoryNew', $parser );
}

# Get more details to create the repository
sub viewAdminRepositoryNewDetails {
	my $parser = new_parser();
	addAdminMenu( $parser );
	$parser->{admin_section_title} = 'Add a Repository';

	# Add the cvsroot
	$parser->{cvsroot} = $fin->{cvsroot};

	# Add the name and label, or details
	$parser->{name} = $fin->{name};
	$parser->{label} = $fin->{label};
	$parser->{details} = $fin->{details} 
		|| "This is a new Repository with the CVSROOT " . (shift || $fin->{cvsroot});

	# Show the page
	page( 'Admin/RepositoryNewDetails', $parser );
}





#####################################################################
# Browse Section Views

# This is the view for the 'Browse Modules' command
sub viewBrowse {
	# If and only if there is only one Module, and it is
	# available, shortcut to it.
	if ( scalar ($MetaData->getModules) == 1
		and ($MetaData->getModules)[0]->available
		and ($MetaData->getModules)[0]->Repository->available
	) {
		$fin->{module} = ($MetaData->getModules)[0]->name;
		return viewBrowseModule();
	}

	# List the Repositories, if there are any
	my $parser = new_parser();
	my @Repositories = $MetaData->getRepositories;
	if ( @Repositories ) {
		$parser->{repositories} = join "\n<br>\n<br>\n",
			map { fragmentBrowseRepository( $_ ) }
			@Repositories;
	} else {
		# If and only if there are no repositories, and no 
		# original explicit command was given to come here,
		# go instead to the about page.
		return viewAbout() unless $fin->{cmd};

		# Show the 'No Repositories' message
		$parser->{repositories} = qq~
			<font color='#990000'>No Repositories have been defined yet for this installation.</font>
			~;
	}

	$gui->{title} = 'Browsing';
	page( 'Browse', $parser );
}

sub viewBrowseModule {
	checkAvailableModule();

	# Get the Activity for the module
	my $Activity = $Module->getActivity
		or Error( "Failed to get Activity for module '$mod'" );

	# Build the Range totals
	my $range_depth;
	if ( defined $fin->{range_depth} ) {
		$range_depth = $fin->{range_depth} eq 'year' ? 'year' : 'month';
	} else {
		my $months = $Activity->months;
		Error( 'Error establishing month range of activity' ) unless defined $months;
		$range_depth = $months > 30 ? 'year' : 'month';
	}

	# Now actually get the range
	my $Range = $Activity->getRangePopulated( 'Basics', $range_depth )
		or Error( "Failed to build totals for module '$mod'" );

	# Extend the range to include today's month
	$Range = $Range->extendToNow()
		or Error( "Failed to extend range to include today's date" );

	# Build the Author totals
	my $AuthorTotals = CVSMonitor::AuthorTotals->new( $Activity )
		or Error( "Failed to get author totals" );

	# Create the parser
	my $parser = new_parser();

	# Add the range section
	$parser->{range} = fragmentRange( $Range );
	if ( $range_depth eq 'year' ) {
		$parser->{range_label} = 'Activity by Year';
		$parser->{range_link} = fragmentLinkBack( 'Show by Month',
			{ range_depth => 'month' }, 
			{ class => '.text_small_note' } 
			);
	} else {
		$parser->{range_label} = 'Activity by Month';
		$parser->{range_link} = fragmentLinkBack( 'Show by Year', 
			{ range_depth => 'year' } 
			);
	}

	# Add the graph
	$parser->{graph} = fragmentGraph( $Module );
	$parser->{graph_selector} = fragmentGraphSelector( $Module );

	# Add the Author sections
	$parser->{author} = fragmentAuthorTotals( $AuthorTotals );

	# Add the activity section
	if ( $fin->{activity} and $fin->{activity} eq 'karma' ) {
		$parser->{activity} = fragmentActivityImportant( $Activity );
		$parser->{activity_label} = 'Important Changes';
		$parser->{activity_link} = fragmentLinkBack( 'Recent Changes', 
			{ activity => 'recent' } );
	} else {
		$parser->{activity} = fragmentActivityRecent( $Activity );
		$parser->{activity_label} = 'Recent Changes';
		$parser->{activity_link} = fragmentLinkBack( 'Important Changes',
			{ activity => 'karma' } );
	}

	# Show the stats page
	$gui->{title} = 'Browsing ' . $Module->name;
	page( 'Browse/Module', $parser );
}

sub viewBrowseRepository {
	checkAvailableRepository();

	# Get the Activity for the repository
	my $Activity = $Repository->getActivity
		or Error( "Failed to get Activity for repository '$rep'" );

	# Work out what depth to show the range totals at
	my $range_depth;
	if ( defined $fin->{range_depth} ) {
		$range_depth = $fin->{range_depth} eq 'year' ? 'year' : 'month';
	} else {
		my $months = $Activity->months;
		Error( 'Error establishing month range of activity' ) unless defined $months;
		$range_depth = $months > 30 ? 'year' : 'month';
	}

	# Now actually get the range
	my $Range = $Activity->getRangePopulated( 'Basics', $range_depth )
		or Error( "Failed to build totals for repository '$rep'" );

	# Extend the range to include today's month
	$Range = $Range->extendToNow()
		or Error( "Failed to extend range to include today's date" );

	# Build the Author totals
	my $AuthorTotals = CVSMonitor::AuthorTotals->new( $Activity )
		or Error( "Failed to get author totals" );

	# Create the parser
	my $parser = new_parser();

	# Add the Range section
	$parser->{range} = fragmentRange( $Range );
	if ( $range_depth eq 'year' ) {
		$parser->{range_label} = 'Activity by Year';
		$parser->{range_link} = fragmentLinkBack( 'Show by Month', {
			range_depth => 'month',
			} );
	} else {
		$parser->{range_label} = 'Activity by Month';
		$parser->{range_link} = fragmentLinkBack( 'Show by Year', {
			range_depth => 'year',
			} );
	}

	# Add the graph
	$parser->{graph} = fragmentGraph( $Repository );
	$parser->{graph_selector} = fragmentGraphSelector( $Repository );

	# Add the authors section
	$parser->{author} = fragmentAuthorTotals( $AuthorTotals );

	# Show the stats page
	$gui->{title} = 'Browsing ' . $Repository->getLabel;
	page( 'Browse/Repository', $parser );
}

sub viewBrowseYear {
	checkAvailableModule();

	# Check the Year
	Error( "No year specified" ) unless $fin->{year};
	my $dateregex = CVSMonitor::DateDepth->seperate_regex( 'year' );
	unless ( $fin->{year} =~ $dateregex ) {
		Error( "Invalid year format" );
	}

	# Get the activity for the module, and filter to the year
	my $Activity = $Module->getActivity
		or Error( "Failed to get Activity" );
	$Activity->filter( 'in', $fin->{year} )
		or Error( "Failed to filter to the year" );

	# Create the parser for the page
	my $parser = new_parser();
	$parser->{year} = $fin->{year};

	# Add the main elements
	if ( $Activity->getVersionCount ) {
		# Create the Range and Author Totals for the Activity
		# and add them to the parser
		my $Range = $Activity->getRangePopulated( 'Basics', 'month' )
			or Error( "Error creating Range" );
		my $AuthorTotals = CVSMonitor::AuthorTotals->new( $Activity )
			or Error( "Failed to get Author Totals for year" );

		$parser->{range} = fragmentRange( $Range );
		$parser->{authors} = fragmentAuthorTotals( $AuthorTotals );
		$parser->{activity} = fragmentActivityImportant( $Activity );
	} else {
		$parser->{range} = "No Activity in this Year";
		$parser->{authors} = "No Activity in this Year";
		$parser->{activity} = "No Activity in this Year";
	}

	# Show the page
	page( 'Browse/Year', $parser );
}

sub viewBrowseMonth {
	checkAvailableModule();

	# Check the Month
	Error( "No month specified" ) unless $fin->{month};
	my $monthregex = CVSMonitor::DateDepth->seperate_regex( 'month' );
	unless ( $fin->{month} =~ $monthregex ) {
		Error( "Invalid month format" );
	}

	# Get the Activity and filter
	my $Activity = $Module->getActivity()
		or Error( "Error getting Activity for Module" );
	$Activity->filter( 'date', $fin->{month} )
		or Error( "Error filtering Activity to month" );

	# Create the Range for the month
	my $Range = CVSMonitor::Range::Basics->month( $fin->{month} )
		or Error( "Error creating month range" );
	$Range->add( $Module ) 
		or Error( "Failed to add to totals for module '$mod'" );

	# Create the Author Totals for the Month
	my $AuthorTotals = CVSMonitor::AuthorTotals->new( $Activity )
		or Error( "Error trying to create Author Totals" );

	# Create the parser and add the basics
	my $parser = new_parser();
	$parser->{month} = $fin->{month};

	# Add the main elements
	if ( $Activity->getVersionCount ) {
		$parser->{range} = fragmentRange( $Range );
		$parser->{authors} = fragmentAuthorTotals( $AuthorTotals );
		$parser->{activity} = fragmentActivityImportant( $Activity );
	} else {
		$parser->{range} = "No Activity in this Year";
		$parser->{authors} = "No Activity in this Year";
		$parser->{activity} = "No Activity in this Year";
	}

	# Show the stats page
	page( 'Browse/Month', $parser );
}

# Show the details ( ChangeSets etc ) for a single day
sub viewBrowseDay {
	checkAvailableModule();

	# Check the Day
	Error( "No day specified" ) unless $fin->{day};
	my $dayregex = CVSMonitor::DateDepth->seperate_regex( 'day' );
	unless ( $fin->{day} =~ $dayregex ) {
		Error( "Invalid day format" );
	}

	# Get the ChangeSets for each and order by date
	my $ChangeSets = $Module->getChangeSets( $fin->{day} ) or return undef;
	@$ChangeSets = sort { $a->dateRaw <=> $b->dateRaw } @$ChangeSets;

	# Create the parser
	my $parser = new_parser();
	$parser->{day} = $fin->{day};

	# Add the month and year for the day
	my $Date = CVSMonitor::DateDepth->new( $fin->{day} );
	$parser->{year} = $Date->year;
	$parser->{month} = $Date->fullMonth;

	# Add the fragment for the Change Sets
	$parser->{change_sets} = fragmentChangeSets( $ChangeSets );

	# Work out the total change for all the ChangeSets
	my @Versions = map { $_->getVersions } @$ChangeSets;
	my $plus = 0;
	my $minus = 0;
	foreach my $Version ( @Versions ) {
		next if $Version->File->binary;
		$plus += $Version->{lines}->{add} + $Version->{lines}->{increase};
		$minus += $Version->{lines}->{remove} + $Version->{lines}->{decrease};
	}
	$plus = $plus ? "+$plus" : '';
	$minus = $minus ? "-$minus" : '';
	$parser->{totalchange} = join ' ', grep { $_ } ( $plus, $minus );

	# Show the page
	page( 'Browse/Day', $parser );
}

sub viewBrowseChangeSet {
	checkAvailableModule();

	# Did they give us an id?
	Error( 'No ChangeSet id provided' ) unless $fin->{id};

	# Try to get the ChangeSet
	my $ChangeSet = $Module->getChangeSet( $fin->{id} );
	unless ( $ChangeSet ) {
		Error( "No ChangeSet found with the id '$fin->{id}'" );
	}

	# Show the page for the ChangeSet
	my $parser = new_parser();
	$parser->{id} = $fin->{id};
	$parser->{changeset} = fragmentChangeSet( $ChangeSet, 'show_all_versions' );
	page( 'Browse/ChangeSet', $parser );
}

# Default file view
# Go straight to the file history
sub viewBrowseFile {
	viewBrowseFileHistory();
}

# Show the file history
sub viewBrowseFileHistory {
	checkAvailableModule();

	# Get the file
	my $File = $Module->getFile( $fin->{file} );
	Error( $Module->errstr ) unless $File;

	# Create the parser and add the basics
	my $parser = new_parser();
	$parser->{file} = $fin->{file};
	$parser->{file_history} = fragmentFileHistory( $File );

	# Add the extra piece of the toolbar
	addToolbar( 'file', $File );

	page( 'Browse/FileHistory', $parser );
}

# Show an annotated version of the file
sub viewBrowseFileAnnotation {
	checkAvailableModule();

	# Get the file
	my $File = $Module->getFile( $fin->{file} );
	Error( $Module->errstr ) unless $File;
	Error( "Cannot show annotation for binary file" ) if $File->binary;

	# Add a linkable module
	my $parser = new_parser();
	$parser->{module} = fragmentLink( $File->Module->name, {
		cmd    => 'viewBrowseModule',
		module => $File->Module->name,
		} );

	# Add the file name
	$parser->{file} = $File->filename;

	# Do we show line numbers?
	$parser->{lines} = fragmentLink( 
		($fin->{lines} ? 'Hide Line Numbers' : 'Show Line Numbers'), 
		{
			cmd    => 'viewBrowseFileAnnotation',
			module => $File->Module->name,
			file   => $fin->{file},
			lines  => ($fin->{lines} ? 0 : 1),
			} );

	# Get the annotation content and escape it
	my $annotation = $File->annotate;
	$annotation =~ s/\t/$TAB/ge;
	$annotation =~ s/\n+$//;
	if ( $fin->{lines} ) {
		$annotation = linenumberify( $annotation );
	}
	$parser->{annotation} = CVSMonitor::HTML->escape( $annotation );

	addToolbar( 'file', $File );
	page( 'Browse/FileAnnotation', $parser );
}

# For the current HEAD version of the file, show a ranking of which
# developers contributed to it.
sub viewBrowseFileBlame {
	checkAvailableModule();

	# Get the file
	my $File = $Module->getFile( $fin->{file} );
	Error( $Module->errstr ) unless $File;
	Error( "Cannot show blame for binary file" ) if $File->binary;

	# Get the blame statistics
	my $blame = CVSMonitor::Statistics->fileBlame( $File );
	Error( 'Error while generating blame statistics' ) unless $blame;

	# Process the blame data
	my @data = ( [ 'Author', 'Lines Contributed' ] );
	my @author_order = sort { $blame->{$b}->{lines} <=> $blame->{$a}->{lines} } keys %$blame;
	foreach ( @author_order ) {
		push @data, [
			fragmentEscapedLink( $_, {
				cmd => 'viewBrowseAuthor',
				module => $File->Module->name,
				author => $_,
				} ),
			commify( $blame->{$_}->{lines} ),
			];
	}

	# Generate the blame table
	my $parser = new_parser();
	$parser->{blame} = CVSMonitor::HTML::Table->table( \@data, {
		top         => 1,
		left        => 1,
		leftcolor   => '#EEEEEE',
		normalalign => 'right',
		altrowcolor => '',
		} );

	# Add the link to annotation view
	$parser->{annotation_link} = fragmentLink( 'Annotation', {
		cmd => 'viewBrowseFileAnnotation',
		module => $File->Module->name,
		file => $File->filename,
		} );

	addToolbar( 'file', $File );
	page( 'Browse/FileBlame', $parser );
}

# Once we get file branch graphing working,
# this is the view for it.
sub viewBrowseFileBranches {
	checkModule();

	# Get the file
	my $File = $Module->getFile( $fin->{file} );
	Error( $Module->errstr ) unless $File;

	# Create the parser and add the basics
	my $parser = new_parser();
	$parser->{file} = $fin->{file};
	$parser->{file_branches} = fragmentFileBranches( $File );

	# Add the file toolbar
	addToolbar( 'file', $File );

	page( 'FileBranches', $parser );
}

# Default to the diff
sub viewBrowseVersion {
	viewBrowseVersionDiff();
}

# Show details for a particular version
sub viewBrowseVersionDiff {
	checkAvailableModule();

	# Get the check the File
	Error( "You did not specify a file name" ) unless $fin->{file};
	my $File = $Module->getFile( $fin->{file} ) or Error( $Module->errstr );

	# Get and check the Version
	Error( "You dod not specift a version" ) unless $fin->{version};
	my $Version = $File->getVersion( $fin->{version} ) or Error( $Module->errstr );

	# Add the date
	my $parser = new_parser();
	$parser->{date} = fragmentDateLink( $Version->date, $Version->Module );

	# Add the link to the author
	my $author = $Version->author;
	if ( $author ) {
		$parser->{author} = fragmentEscapedLink( $author, {
			cmd => 'viewBrowseAuthor',
			module => $Version->Module->name,
			author => $author,
			} );
	} else {
		$parser->{author} = 'Unknown Author';
	}

	# Add the link to the file
	$parser->{file} = fragmentLink( $File->filename, {
		cmd => 'viewBrowseFile',
		module => $Version->Module->name,
		file => $File->filename,
		} );

	# Add the revision basics
	$parser->{revision} = $Version->revision;
	$parser->{change} = $Version->changeString;
	$parser->{comment} = CVSMonitor::HTML->escape_multiline( $Version->comment );
	$parser->{comment} = linkToTrackers( $Version->Module, $parser->{comment} ) || $parser->{comment};

	# Add the diff content
	if ( $Version->File->binary ) {
		$parser->{diff} = "No diffs available for binary files";
	} elsif ( ! $Version->previous ) {
		$parser->{diff} = "This is a newly created File. No previous version to diff from";
	} else {
		my $diff = $Version->diff;
		if ( defined $diff ) {
			$parser->{diff} = CVSMonitor::HTML->escape( $diff );
		} else {
			$parser->{diff} = "Failed to get diff for version";
		}
	}

	# Add the link to the ChangeSet
	my $ChangeSet = $Version->ChangeSet();
	if ( $ChangeSet ) {
		$parser->{changeset} = fragmentLink( 'ChangeSet ' . $ChangeSet->id, {
			cmd => 'viewBrowseChangeSet',
			module => $Version->Module->name,
			id => $ChangeSet->id
			} );
	} else {
		$parser->{changeset} = 'Unknown ChangeSet';
	}

	# Show the page
	addToolbar( 'file', $File );
	addToolbar( 'version', $Version );
	$gui->{title} = "Version $parser->{revision} Diff";
	page( 'Browse/Version', $parser );
}

sub viewBrowseAuthor {
	checkAvailableModule();

	# Check the author
	$fin->{author} =~ s/\s//g;
	Error( "You did not specify an author name" ) unless length $fin->{author};
	unless ( $MetaData->_checkFormat( 'username', $fin->{author} ) ) {
		Error( 'Illegal author name format' );
	}

	# Get the author
	my $Author = $Module->getAuthor( $fin->{author} )
		or Error( "Could not find Author '" . CVSMonitor::HTML->escape($fin->{author}) . "'" );

	# Get the Activity for the Module and limit to the author
	my $Activity = $Module->getActivity
		or Error( "Error while getting Activity for Module" );
	$Activity->filter( 'author', $Author->name )
		or Error( "Error while attempting to filter to the author" );

	# Is there anything for that author
	unless ( $Activity->getVersionCount ) {
		Error( "The author either does not exist in the repository, or has never made a commit" );
	}

	# Create the parser and add the basics
	my $parser = new_parser();
	$parser->{author} = CVSMonitor::HTML->escape( $Author->name );

	# Build the Range totals
	my $range_depth;
	if ( defined $fin->{range_depth} ) {
		$range_depth = $fin->{range_depth} eq 'year' ? 'year' : 'month';
	} else {
		my $months = $Activity->months;
		Error( 'Error establishing month range of activity' ) unless defined $months;
		$range_depth = $months > 30 ? 'year' : 'month';
	}

	# Now actually get the range, extend to include today,
	# and add the activity for the Author
	my $Range = $Activity->getRange( 'Basics', $range_depth )
		or Error( "Failed to build totals for module '$mod'" );
	$Range = $Range->extendToNow or Error( "Failed to extend range to include today's date" );
	$Range->add( $Activity ) or Error( "Failed to add author's Activity to Range" );

	# Add the Range section
	$parser->{range} = fragmentRange( $Range );
	if ( $range_depth eq 'year' ) {
		$parser->{range_label} = 'Activity by Year';
		$parser->{range_link} = fragmentLinkBack( 'Show by Month', {
			range_depth => 'month',
			} );
	} else {
		$parser->{range_label} = 'Activity by Month';
		$parser->{range_link} = fragmentLinkBack( 'Show by Year', {
			range_depth => 'year',
			} );
	}

	# Add the graph
	$parser->{graph} = fragmentGraph( $Author );
	$parser->{graph_selector} = fragmentGraphSelector( $Author );

	# Add the Activity section
	if ( $fin->{activity} and $fin->{activity} eq 'karma' ) {
		$parser->{activity} = fragmentActivityImportant( $Activity );
		$parser->{activity_label} = 'Important Changes';
		$parser->{activity_link} = fragmentLinkBack( 'Recent Changes', 
			{ activity => 'recent' } );
	} else {
		$parser->{activity} = fragmentActivityRecent( $Activity );
		$parser->{activity_label} = 'Recent Changes';
		$parser->{activity_link} = fragmentLinkBack( 'Important Changes',
			{ activity => 'karma' } );
	}

	page( 'Browse/Author', $parser );
}

sub rssModule {
	checkAvailableModule();

	# Generate the RSS XML
	my $Generator = XML::Generator->new(
		escape => 'always',
		pretty => 2,
		);
	my $rss = $Generator->rss(
		{ version => '2.0' },
		rssModuleChannel( $Generator, $Module ),
		);

	# Send the rss document
	print CGI::header( 'text/xml' ) . $rss;

	1;
}

sub rssModuleChannel {
	my $Generator = shift;
	my $module = shift;

	my $module_url = CGI::url()
		. "?cmd=viewBrowseModule"
	        . "&module=" . $module->name;

	$Generator->channel(
		$Generator->title( $module->name . " CVS Activity" ),
		$Generator->link( $module_url ),
		$Generator->description( "CVS activity for " . $module->name ),
		$Generator->generator( "CVS Monitor $versionstring" ),
		rssModuleItems( $Generator, $module )
		);
}

sub rssModuleItems {
	my $Generator = shift;
	my $module = shift;

	require Date::Format;

	my @items;
	my $ChangeSets = $module->getActivity->getRecentChangeSets( $Config->getGuiChangesets );
	foreach my $ChangeSet ( @$ChangeSets ) {
		my $title = $module->name . " ChangeSet " . $ChangeSet->id;
		my $changeset_url = CGI::url()
			. "?cmd=viewBrowseChangeSet"
			. "&module=" . $module->name
			. "&id=" . $ChangeSet->id;

		push @items, $Generator->item(
			$Generator->title( $title ),
			$Generator->description( $ChangeSet->comment ),
			$Generator->author( $ChangeSet->author ),
			$Generator->pubDate( Date::Format::time2str(
				"%a, %e %b %Y %X %Z", $ChangeSet->dateRaw
				) ),
			$Generator->link( $changeset_url ),
			$Generator->guid( $changeset_url ),
			);
	}

	@items;
}

sub viewGrapher {
	# Get the type of graph
	if ( $fin->{source} eq 'module' ) {
		viewGrapherModule();
	} elsif ( $fin->{source} eq 'author' ) {
		viewGrapherAuthor();
	} elsif ( $fin->{source} eq 'repository' ) {
		viewGrapherRepository();
	} else {
		Error( "Unknown or missing grapher type" );
	}
}

sub viewGrapherRepository {
	checkAvailableRepository();

	# Get a new parser
	my $parser = new_parser();

	# Generate a graph image for the module
	$parser->{graph} = fragmentGraph( $Repository, $fin->{type}, 'grapher' );

	page( 'Grapher', $parser );
}

sub viewGrapherModule {
	checkAvailableModule();

	# Get a new parser
	my $parser = new_parser();

	# Generate a graph image for the module
	$parser->{graph} = fragmentGraph( $Module, $fin->{type}, 'grapher' );

	page( 'Grapher', $parser );
}

sub viewGrapherAuthor {
	checkAvailableModule();

	# Check the author
	$fin->{author} =~ s/\s//g;
	Error( "You did not specify an author name" ) unless length $fin->{author};
	unless ( $MetaData->_checkFormat( 'username', $fin->{author} ) ) {
		Error( 'Illegal author name format' );
	}

        # Get the author
        my $Author = $Module->getAuthor( $fin->{author} )
                or Error( "Could not find Author '" . CVSMonitor::HTML->escape($fin->{author}) . "'" );

	# Get a new parser
	my $parser = new_parser();
	$parser->{author} = CVSMonitor::HTML->escape( $Author->name );

	# Generate the graph image
	$parser->{graph} = fragmentGraph( $Author, $fin->{type}, 'grapher' );

	page( 'Grapher', $parser );
}

sub viewConfigMissing {
	# Don't show the toolbar
	addToolbar( 'main', 'disabled' );

	# Just create a parser and show the page
	page( "Error/ConfigMissing", new_parser() );
}

sub viewConfigEdit {
	# To view this, either they must be authenticated, or
	# there is no Config file
	if ( $Config ) {
		unless ( $Config->isNew or authenticate() ) {
			return viewAdminLogin();
		}
	} else {
		# Set the Config to a new Config
		$Config = CVSMonitor::Config->new()
			or Error( "Error creating new config" );
	}

	# If the Config is new, try to import from setup.conf
	if ( $Config->isNew ) {
		# Don't show the entire menu bad the first time
		addToolbar( 'main', 'disabled' );

		if ( $Config->importSetupValues ) {
			$message = 'Imported values from setup process correctly';
		} else {
			Error( 'Failed to import values from setup process correctly<br>'
				. $Config->errstr );
		}

		# Re-initialise the gui
		initialise_gui();
	}

	# Add the parser items
	my $parser = new_parser();
	$parser->{Config_metadata_directory} = $fin->{metadata_directory}
		|| $Config->getHostScratch();
	$parser->{Config_admin_password} = $fin->{admin_password}
		|| $Config->getSecurityAdminpassword();
	$parser->{Config_gui_cachefilesystem} = $fin->{gui_cachefilesystem}
		|| $Config->getHostStaticfilesystem();
	$parser->{Config_gui_cachehttp} = $fin->{gui_cachehttp}
		|| $Config->getHostStatichttp();
	$parser->{Config_gui_changesets} = $fin->{gui_changesets}
		|| $Config->getGuiChangesets();

	# Check the appropriate radio button for the time format
	$parser->{"checked_" . ($fin->{gui_timeformat} || $Config->getGuiTimeformat) } = 'checked';

	# Get the user we are executing as and show it to them
	### FIXME - Replace with a better way
	$parser->{web_user} = `whoami`;
	chomp $parser->{web_user};

	# If they are using debian, and the working directory is below /tmp,
	# show the additional message.
	if ( $Config->getHostScratch =~ m!/tmp/?! ) {
		if ( -e '/etc/debian_version' ) {
			$parser->{debian_note} = qq~
				Note To Debian Users<br>
				It appears you are using Debian, and the working directory
				is currently set to somewhere under /tmp. Please note that Debian
				erases /tmp on boot, and has a dedicated cache directory of /var/cache 
				that should be used inplace of /tmp. You can create a cache
				directory for CVS Monitor in the prefered location with the 
				following commands, which will probably require root access.<br><br>
				mkdir /var/cache/cvsmonitor<br>
				chown -R www-data:www-data /var/cache/cvsmonitor<br><br>
				I will not prevent you from using /tmp, but you should be aware 
				of the loss of the config/cache should this server ever reboot.
				~;
		}
	}

	page( "Admin/ConfigEdit", $parser );
}

sub viewSearch {
	# Shortcut to the search space page
	$fin->{cmd} = 'viewSearchChangeSet';

	# Add the search toolbar
	addToolbar( 'search' );

	# If there are no modules at all within which to search, so an error message
	my $available = 0;
	foreach ( $MetaData->getModules ) {
		$available++ if $_->available;
	}
	unless ( $available ) {
		Error( "There are no modules available in which to search" );
	}

	page( 'Search', new_parser() );
}

sub viewSearchSpace {
	# Add the search toolbar
	addToolbar( 'search' );

	# Create a parser
	my $parser = new_parser();
	$parser->{cmd} = $fin->{cmd};

	# Generate the search space selector if there is something to search
	if ( $MetaData->getRepositories ) {
		my @rows = ( [ 'Repository', 'Modules' ] );
		foreach my $Repository ( $MetaData->getRepositories ) {
			my $count = 0;
			my @module_links = ();
			foreach my $Module ( $Repository->getModules ) {
				if ( $Module->available ) {
					push @module_links, fragmentLinkBack( $Module->display, {
						space => $Module->name,
						} );
					$count++;
				} else {
					push @module_links, fragmentLinkDisabled( $Module->display );
				}
			}

			# Add the completed data to the rows
			push @rows, [ ($count 
				? fragmentLinkBack( $Repository->getLabel, {
					space => $Repository->name,
					} )
				: fragmentLinkDisabled( $Repository->getLabel )),
				join( ' &middot; ', @module_links )
				];
		}
		$parser->{search_space} = CVSMonitor::HTML::Table->table( \@rows, { 
			top         => 1,
			topclass    => 'table_header',
			} );
	} else {
		$parser->{search_space} = qq~
			<font color='#990000'>No Repositories have been defined yet for this installation.<br>
			There is nothing to search on.</font>
			~;
	}

	page( 'Search/Space', $parser );
}

sub viewSearchChangeSet {
	$gui->{title} = 'ChangeSet Search';

	# Try to find the search space
	my $Activity = findSearchSpace();
	return viewSearchSpace() unless $Activity;

	# Add the search toolbar
	addToolbar( 'search' );

	# Create the parser
	my $parser = new_parser();
	$parser->{space} = $fin->{space};

	# Add the comment options
	$parser->{search_message} = $fin->{search_message} if $fin->{search_message};
	$parser->{search_message_type} = CVSMonitor::HTML::Form->dropbox( 'search_message_type', [
			[ 'substring', 'Literal Substring' ],
			[ 'regex', 'Regular Expression' ],
			], 
		($fin->{search_message_type} || 'literal'),
		{ notnull => 1 }
		);
	$parser->{search_message_case} = CVSMonitor::HTML::Form->dropbox( 'search_message_case', [
			[ '1', 'Case Sensitive' ],
			[ '0', 'Case In-sensitive' ],
			],
		$fin->{search_message_case} || '0',
		{ notnull => 1, }
		);

	# Find the authors in the search space
	my $authors = $Activity->getAuthorNames
		or Error( 'Failed to get author names in search space' );
	$authors = [ map { [ $_, $_ ] } sort @$authors ];

	# The dropbox generator should auto-escape special usernames
	$parser->{search_author} = CVSMonitor::HTML::Form->dropbox( 'search_author', $authors, $fin->{search_author}, {
		nullstring => 'Any Author' } );

	# Generate the sort by radio button cluster
	$parser->{search_sortby} = CVSMonitor::HTML::Form->radio( 'search_sortby', [
			[ 'karma', 'Size ( As determined by Karma )' ],
			[ 'newest', 'Date (newest to oldest)' ],
			[ 'oldest', 'Date (oldest to newest)' ],
			],
		$fin->{search_sortby} || 'karma',
		'vertical' );

	page( 'Search/ChangeSet', $parser );
}

sub viewSearchChangeSetResults {
	my $Activity = shift;
	unless ( isa( $Activity, 'CVSMonitor::MetaData::Activity' ) ) {
		Error( 'Unable to view ChangeSets. We were not passed any' );
	}

	# What range of ChangeSets do we want to view
	my $window_start = $fin->{window_start} || 0;
	my $window_length = $fin->{window_length} || 20;

	# Create a Window to do the math
	my $Window = Array::Window->new( source => $Activity,
		window_start => $window_start,
		window_length => $window_length );
	Error( "Math object failed while calculating which ChangeSets to show" ) unless $Window;

	# Create the parser
	my $parser = new_parser();

	# Get the ChangeSets to show, and the html for them
	if ( $Window->required ) {
		my @ChangeSets = $Activity->getElementRange( 
			($Window->window_start + 1),
			($Window->window_end + 1) )
			or Error( "Failed to select subset of ChangeSets results to display" );
		$parser->{results} = fragmentChangeSets( \@ChangeSets );

		# Create the nav links
		my $PreviousWindow = $Window->previous;
		my $previous_link = $PreviousWindow
			? fragmentLinkBack( "Previous $window_length", {
				window_start => $PreviousWindow->window_start,
				window_length => $window_length,
				} )
			: fragmentLinkDisabled( "Previous $window_length" );
		my $NextWindow = $Window->next;
		my $next_link = $NextWindow
			? fragmentLinkBack( "Next $window_length", {
				window_start => $NextWindow->window_start,
				window_length => $window_length,
				} )
			: fragmentLinkDisabled( "Next $window_length" );
		$parser->{navigation} = "$previous_link $next_link";

		# Add the results comment
		$parser->{comment} = "Showing ChangeSets " 
			. ($Window->window_start + 1)
			. " to "
			. ($Window->window_end + 1)
			. " of the "
			. $Activity->elements
			. " ChangeSets that matched your request";

	} else {
		# No Navigation, just generate for the Activity
		$parser->{results} = fragmentChangeSets( $Activity );

		# Add the results comment
		$parser->{comment} = "Found " 
			. $Activity->elements 
			. " ChangeSets that matched your request";

	}


	# Show the page
	addToolbar( 'search' );
	$gui->{title} = 'ChangeSet Search Results';
	page( 'Search/ChangeSetResults', $parser );
}

sub viewSearchFile {
	$gui->{title} = 'File Search';

	# Try to find the search space
	my $Activity = findSearchSpace();
	return viewSearchSpace() unless $Activity;

	# Add the search toolbar
	addToolbar( 'search' );

	# Create the parser
	my $parser = new_parser();
	$parser->{space} = $fin->{space};

	# Add the filename search
	$parser->{search_filename} = $fin->{search_filename} if $fin->{search_filename};
	$parser->{search_filename_type} = CVSMonitor::HTML::Form->dropbox( 'search_filename_type', [
			[ 'exact', 'Exact Filename' ],
			[ 'file.substring', 'Filename Substring' ],
			[ 'path.substring', 'Path Substring' ],
			[ 'regex', 'Path Regular Expression' ],
			],
		$fin->{search_filename_type} || 'file.substring',
		{ notnull => 1 }
		);
	$parser->{search_filename_case} = CVSMonitor::HTML::Form->dropbox( 'search_filename_case', [
			[ '1', 'Case Sensitive' ],
			[ '0', 'Case In-sensitive' ],
			],
		$fin->{search_filename_case} || '0',
		{ notnull => 1, }
		);

	# Find the authors in the search space
	my $authors = $Activity->getAuthorNames
		or Error( 'Failed to get author names in search space' );
	$authors = [ map { [ $_, $_ ] } sort @$authors ];
	$parser->{search_author} = CVSMonitor::HTML::Form->dropbox( 'search_author', $authors, $fin->{search_author}, {
		nullstring => 'Any Author' } );

	# Add the 'exists' condition
	$parser->{search_exists} = CVSMonitor::HTML::Form->checkbox( 'search_exists', 1, { 
		checked => (exists $fin->{search_filename} ? $fin->{search_exists} : 1)
		} );

	# Generate the sort by radio button cluster
	$parser->{search_sortby} = CVSMonitor::HTML::Form->radio( 'search_sortby', [
			[ 'path', 'Full File Path' ],
			[ 'file', 'Filename' ],
			#[ 'newest', 'Creation Date (newest to oldest)' ],
			#[ 'oldest', 'Creation Date (oldest to newest)' ],
			],
		($fin->{search_sortby} || 'path'),
		'vertical' );

	page( 'Search/File', $parser );
}

sub viewSearchFileResults {
	my $Files = shift;
	unless ( isa( $Files, 'ARRAY' ) and isa( $Files->[0], 'CVSMonitor::MetaData::File' ) ) {
		Error( 'Unable to view Files. We were not passed any' );
	}

	# What range of ChangeSets do we want to view
	my $window_start = $fin->{window_start} || 0;
	my $window_length = $fin->{window_length} || 50;

	# Create a Window to do the math
	my $Window = Array::Window->new( source => $Files,
		window_start => $window_start,
		window_length => $window_length );
	Error( "Math object failed while calculating which Files to show" ) unless $Window;

	# Do we need to show the module the file was in?
	my $opt = { module => $fin->{space} =~ /\./ ? 0 : 1 };

	# Create the parser
	my $parser = new_parser();

	# Get the ChangeSets to show, and the html for them
	my $file_count = scalar @$Files;
	if ( $Window->required ) {
		@$Files = @$Files[($Window->window_start + 1) .. ($Window->window_end + 1)];
		$parser->{results} = fragmentFiles( $Files, $opt );

		# Create the nav links
		my $PreviousWindow = $Window->previous;
		my $previous_link = $PreviousWindow
			? fragmentLinkBack( "Previous $window_length", {
				window_start => $PreviousWindow->window_start,
				window_length => $window_length,
				} )
			: fragmentLinkDisabled( "Previous $window_length" );
		my $NextWindow = $Window->next;
		my $next_link = $NextWindow
			? fragmentLinkBack( "Next $window_length", {
				window_start => $NextWindow->window_start,
				window_length => $window_length,
				} )
			: fragmentLinkDisabled( "Next $window_length" );
		my $start_text = $Window->window_start + 1;
		my $end_text = $Window->window_end + 1;
		$parser->{navigation} = "$previous_link <-- $start_text - $end_text --> $next_link";

		# Add the results comment
		$parser->{comment} = "Showing Files $start_text to $end_text of the $file_count Files that matched your request";

	} else {
		# No Navigation, just generate for the Activity
		$parser->{results} = fragmentFiles( $Files, $opt );

		# Add the results comment
		$parser->{comment} = "Found $file_count Files that matched your request";
	}

	# Show the page
	addToolbar( 'search' );
	$gui->{title} = 'File Search Results';
	page( 'Search/FileResults', $parser );
}





#####################################################################
# Graph Generators

sub graphRepositoryLines {
	checkAvailableRepository( 'die' );

	my $display = $Repository->getLabel;
	graphRepository( 'Lines', "Repository '$display' Growth" );
}

sub graphRepositoryKarma {
	checkAvailableRepository( 'die' );

	my $display = $Repository->getLabel;
	graphRepository( 'Karma', "Repository '$display' Karma" );
}

sub graphRepositoryFiles {
	checkAvailableRepository( 'die' );

	my $display = $Repository->getLabel;
	graphRepository( 'Files', "Repository '$display' Growth" );
}

sub graphRepository {
	my $type = shift || 'Lines';
	my $title = shift;

	# Get the totals
	my $Range = $Repository->getRangePopulated( ($type eq 'Files' ? 'Files' : 'Basics'), 'month' )
		or die( "Failed to build totals for module '$mod'" );
	$Range = $Range->extendToNow() or die( "Failed to extend to current month" );

	# Create a Graph::Range for the Range
	my $graph = CVSMonitor::Graph->new( $type, $Range )
		or die( "Failed to create graph" );
	$graph->setTitle( $title ) or die( "Failed to set title" );
	$graph->setHeight( $fin->{height} ) if $fin->{height};
	$graph->setWidth( $fin->{width} ) if $fin->{width};

	# Allow passing through of the maximum intervals value
	if ( $fin->{maximum_intervals} ) {
		$graph->setMaximumIntervals( $fin->{maximum_intervals} );
	}

	$graph;
}

sub graphModuleLines {
	checkAvailableModule( 'die' );

	my $display = $Module->display;
	graphModule( 'Lines', "Module '$display' Growth" );
}

sub graphModuleKarma {
	checkAvailableModule( 'die' );

	my $display = $Module->display;
	graphModule( 'Karma', "Module '$display' Karma" );
}

sub graphModuleFiles {
	checkAvailableModule( 'die' );

	my $display = $Module->display;
	graphModule( 'Files', "Module '$display' Growth" );
}

sub graphModule {
	my $type = shift || 'Lines';
	my $title = shift;

	# Get the totals
	my $Range = $Module->getRangePopulated( ($type eq 'Files' ? 'Files' : 'Basics'), 'month' )
		or die( "Failed to build totals for module '$mod'" );
	$Range = $Range->extendToNow() or die( "Failed to extend to current month" );

	# Create a Graph::Range for the Range
	my $graph = CVSMonitor::Graph->new( $type, $Range )
		or die( "Failed to create graph" );
	$graph->setTitle( $title ) or die( "Failed to set title" );
	$graph->setHeight( $fin->{height} ) if $fin->{height};
	$graph->setWidth( $fin->{width} ) if $fin->{width};

	# Allow passing through of the maximum intervals value
	if ( $fin->{maximum_intervals} ) {
		$graph->setMaximumIntervals( $fin->{maximum_intervals} );
	}

	$graph;
}

sub graphAuthorLines {
	checkAvailableModule( 'die' );

	# Check the author
	$fin->{author} =~ s/\s//g;
	Error( "You did not specify an author name" ) unless length $fin->{author};
	unless ( $MetaData->_checkFormat( 'username', $fin->{author} ) ) {
		Error( 'Illegal author name format' );
	}

	my $module = $Module->display;
	graphAuthor( 'Lines', "Author '$fin->{author}' Activity" );
}

sub graphAuthorKarma {
	checkAvailableModule( 'die' );

	# Check the author
	$fin->{author} =~ s/\s//g;
	Error( "You did not specify an author name" ) unless length $fin->{author};
	unless ( $MetaData->_checkFormat( 'username', $fin->{author} ) ) {
		Error( 'Illegal author name format' );
	}

	my $module = $Module->display;
	graphAuthor( 'Karma', "Author '$fin->{author}' Karma" );
}

sub graphAuthorFiles {
	checkAvailableModule( 'die' );

	# Check the author
	$fin->{author} =~ s/\s//g;
	Error( "You did not specify an author name" ) unless length $fin->{author};
	unless ( $MetaData->_checkFormat( 'username', $fin->{author} ) ) {
		Error( 'Illegal author name format' );
	}

	my $module = $Module->display;
	graphAuthor( 'Files', "Author '$fin->{author}' Growth" );
}

sub graphAuthor {
	my $type = shift || 'Lines';
	my $title = shift;

	# Get the Activity for the Module and limit to the author
	my $Activity = $Module->getActivity
		or die( "Error while getting Activity for Module" );
	$Activity->filter( 'author', $fin->{author} )
		or die( "Error while attempting to filter to the author" );

	# Is there anything for that author
	unless ( $Activity->getVersionCount ) {
		die( "The author either does not exist in the repositary, or has never made a commit" );
	}

	# Create a Range for the entire module, and add the
	# Activity for the author to it.
	my $Range = $Module->getRange( ($type eq 'Files' ? 'Files' : 'Basics'), 'month' )
		or Error( "Failed to create Range for Module" );
	$Range->add( $Activity )
		or Error( "Failed to add author's Activity to Range" );

	# Create a graph for the Range
	my $graph = CVSMonitor::Graph->new( $type, $Range )
		or die( "Failed to create graph" );
	$graph->setHeight( $fin->{height} ) if $fin->{height};
	$graph->setWidth( $fin->{width} ) if $fin->{width};

	# Allow passing through of the maximum intervals value
	if ( $fin->{maximum_intervals} ) {
		$graph->setMaximumIntervals( $fin->{maximum_intervals} );
	}

	$graph;
}

# Get the graph options for a variety of objects
sub graphOptions {
	my $Object = shift or return undef;
	my $grapher = shift;

	# Split on the object type and create what we need
	my @options = ();
	if ( isa( $Object, 'CVSMonitor::MetaData::Module' ) ) {
		# Options for a Module
		foreach my $type ( qw{Lines Files} ) {
			my $image = graphDynamicImageSrc( $type, $Object, $grapher );
			my $href = graphGrapherLinkHref( $type, $Object );
			push @options, {
				image => $image,
				href  => $href,
				label => $type,
				};
		}

	} elsif ( isa( $Object, 'CVSMonitor::MetaData::Repository' ) ) {
		# Options for a repository
		foreach my $type ( qw{Lines Files} ) {
			my $image = graphDynamicImageSrc( $type, $Object, $grapher );
			my $href = graphGrapherLinkHref( $type, $Object );
			push @options, {
				image => $image,
				href  => $href,
				label => $type,
				};
		}

	} elsif ( isa( $Object, 'CVSMonitor::MetaData::Author' ) ) {
		# Options for an author, in a module
		foreach my $type ( qw{Lines Files Karma} ) {
			my $image = graphDynamicImageSrc( $type, $Object, $grapher );
			my $href = graphGrapherLinkHref( $type, $Object );
			push @options, {
				image => $image,
				href  => $href,
				label => $type,
				};
		}

	} else {
		return undef;
	}

	\@options;
}

sub graphDynamicImageSrc {
	my $type = shift || 'Lines';
	my $Source = shift or return undef;
	my $grapher = shift;

	# Create the basic hash
	my $opt = {
		width             => $grapher ? 800 : 400,
		height            => $grapher ? 600 : 300,
		maximum_intervals => $grapher ? 20 : 10,
		};

	# Determine the reference
	if ( isa( $Source, 'CVSMonitor::MetaData::Module' ) ) {
		$opt->{cmd} = "graphModule$type";
		$opt->{module} = $Source->name;

	} elsif ( isa( $Source, 'CVSMonitor::MetaData::Repository' ) ) {
		$opt->{cmd} = "graphRepository$type";
		$opt->{repository} = $Source->name;

	} elsif ( isa( $Source, 'CVSMonitor::MetaData::Author' ) ) {
		$opt->{cmd} = "graphAuthor$type";
		$opt->{author} = $Source->name;
		$opt->{module} = $Module->name;

	} else {
		return undef;
	}

	# Create and return the image href
	"$callback?" . join( '&', map { $_ . '=' . URI::Escape::uri_escape($opt->{$_}) } keys %$opt );
}

sub graphGrapherLinkHref {
	my $type = shift || 'Lines';
	my $Source = shift;

	# Create the basic hash
	my $opt = {
		cmd               => 'viewGrapher',
		type              => $type,
		maximum_intervals => 20,
		};

	# Determine the reference
	if ( isa( $Source, 'CVSMonitor::MetaData::Module' ) ) {
		$opt->{module} = $Source->name;
		$opt->{source} = 'module';

	} elsif ( isa( $Source, 'CVSMonitor::MetaData::Repository' ) ) {
		$opt->{repository} = $Source->name;
		$opt->{source} = 'repository';

	} elsif ( isa( $Source, 'CVSMonitor::MetaData::Author' ) ) {
		$opt->{author} = $Source->name;
		$opt->{module} = $Module->name;
		$opt->{source} = 'author';

	} else {
		return undef;
	}

	# Return the generated href
	"$callback?" . join( '&', map { $_ . '=' . URI::Escape::uri_escape($opt->{$_}) } keys %$opt );
}





#####################################################################
# Generated HTML Fragments

# Create a generic callback link.
# Don't escape, that is left to the caller
sub fragmentLink {
	my $text = shift;
	my $args = shift || {};
	return undef unless isa( $args, 'HASH' );
	my $options = shift || {};
	return undef unless isa( $options, 'HASH' );

	# Copy the options and args to avoid 
	# corrupting passed values
	$args = { %$args };
	$options = { %$options };

	# Avoid warnings in the map below
	foreach ( keys %$args ) {
		$args->{$_} = '' unless defined $args->{$_};
	}

	# Create the href
	my $arg_string = join( '&', map { "$_=$args->{$_}" } keys %$args );
	my $href = length($arg_string) ? "$callback?$arg_string" : $callback;

	# Create the full set of tag options
	$options->{href} = $href;
	CVSMonitor::HTML->tag_pair( 'a', $options, $text );
}

# As above, but escape automatically
sub fragmentEscapedLink {
	my $text = shift;
	my $args = shift || {};

	# Escape the arguments
	foreach ( keys %$args ) {
		$args->{$_} = URI::Escape::uri_escape( $args->{$_} );
	}

	# Hand off to the normal link generator
	fragmentLink( $text, $args, @_ );
}

# Create a link that is the same as the current one, with some changes
sub fragmentLinkBack {
	my $text = shift;
	my $changes = shift || {};
	my $options = shift;

	# Copy the current options
	my $args = { %$fin };

	# Apply the changes
	foreach ( keys %$changes ) {
		if ( defined $changes->{$_} ) {
			$args->{$_} = $changes->{$_};
		} else {
			delete $args->{$_};
		}
	}

	# Hand off to the normal link generator
	fragmentLink( $text, $args, $options );
}

# This is a silly little sub to wrap the text in a grey font
sub fragmentLinkDisabled {
	my $text = shift;
	"<span class='text_subtle'>$text</span>";
}

# Render a Range to HTML
sub fragmentRange {
	my $Range = shift;
	my $name = shift || $mod;
	return undef unless isa( $Range, 'CVSMonitor::Range::Basics' );

	# Transform to HTML
	my $depth = $Range->depth;
	my $command = {
		year  => 'viewBrowseYear',
		month => 'viewBrowseMonth',
		day   => 'viewBrowseDay',
		}->{$depth};

	# Create the field data
	my @data = ( [ ucfirst lc $depth, qw{Karma Commits Added Removed Change} ],
		map { [
			($_->{commits} && $fin->{module}
				? fragmentLink( $_->{date}, {
					cmd => $command,
					module => $name,
					$depth => $_->{date},
					} )
				: $_->{date}),
			commify( $_->{karma} ),
			commify( $_->{commits} ), 
			commify( $_->{increase} + $_->{add} ), 
			commify( $_->{decrease} + $_->{remove} ),
			commify( $_->{increase} + $_->{add} - $_->{decrease} - $_->{remove} ),
			] } @$Range,
		);

	# Add the totals
	my $totals = $Range->totals;
	foreach ( keys %$totals ) {
		$totals->{$_} = commify( $totals->{$_} );
	}
	push @data, [ "Totals", 
		$totals->{karma},
		$totals->{commits}, 
		$totals->{up},
		$totals->{down},
		$totals->{change},
		];

	CVSMonitor::HTML::Table->table( \@data, { 
		top         => 1,
		topclass    => 'table_header',
		bottom      => 1,
		outerwidth  => 0,
		normalalign => 'right',
		} );
}

# Generates the html for the row for a single module
sub fragmentAdminModuleRow {
	my $Module = isa( $_[0], 'CVSMonitor::MetaData::Module' )
		? shift : return undef;

	# Pre-generate some strings we'll need later
	my $name = $Module->name;
	my $display = $Module->display;
	my $state = $Module->state;
	my $date = fragmentEventTime( $Module->dateRaw ) || '&nbsp';
	my $footprint = $Module->footprint;
	$footprint = (defined $footprint and $footprint > 0)
		? AppLib::String->humanFileSize( $footprint )
		: '&nbsp;';

	my $updateTime = $Module->updateTime;
	$updateTime = defined $updateTime
		? (' &middot; ' . Time::Duration::duration( $updateTime, 2 ))
		: '';
	my $updateCost = $Module->updateCost;
	$updateCost = (defined $updateCost and $updateCost > 0)
		? AppLib::String->humanFileSize( $updateCost )
		: 'Unknown';

	# Create all the possible links
	my %links = (
		details => fragmentLink( 'Details', {
			cmd => 'viewAdminModuleDetails',
			module => $name,
			} ),
		browse => fragmentLink( 'Browse', {
			cmd => 'viewBrowseModule',
			module => $name,
			} ),
		checkout => fragmentLink( 'Checkout', {
			cmd => 'adminBackgroundCheckout',
			module => $name,
			} ),
		update => fragmentLink( 'Update', {
			cmd => 'adminBackgroundUpdate',
			module => $name,
			} ),
		release => fragmentLink( 'Release', {
			cmd => 'adminRelease',
			module => $name,
			} ),
		'delete' => fragmentLink( 'Delete', {
			cmd => 'adminDeleteModule',
			module => $name,
			} ),
		recover => fragmentLink( 'Recover', {
			cmd => 'adminRecover',
			module => $name,
			} ),
		errorlog => fragmentLink( 'Error Log', {
			cmd => 'viewAdminModuleErrorLog',
			module => $name,
			} ),
		);

	# Set the available commands based on state
	my $command = '&nbsp';
	if ( $state eq 'Unavailable' ) {
		$command = "$links{details} &middot; $links{checkout} &middot; $links{delete}";
	} elsif ( $Module->transient ) {
		# If the checkout has died, offer a recover option
		if ( $Module->broken ) {
			$command = "$links{details} &middot; $links{recover} &middot; $links{errorlog}";
		} elsif ( $state eq 'Updating' ) {
			$command = "$links{details} &middot; $links{browse}";
		} else {
			$command = "$links{details}";
		}
	} elsif ( $state eq 'Available' ) {
		# Module is in a normal, checked out, state
		$command = "$links{details} &middot; $links{browse} &middot; $links{update} &middot; $links{release}";
	} else {
		Error( "Module '$name' appears to be in an invalid state" );
	}

	# Escape the module display name
	$display = CVSMonitor::HTML->escape( $display );

	# Establish the background colour
	my $color = fragmentModuleColor( $Module );

	# Add more information about the activeness of the update or checkout
	if ( $Module->transient and ! $Module->broken ) {
		$state = "<b>$state...</b>";
	}

	# Add the row
	return qq~
	  <tr> 
	    <td nowrap bgcolor='$color'><b>$display</b></td>
	    <td nowrap bgcolor='$color'>$state</td>
	    <td nowrap bgcolor='$color'>$date</td>
	    <td nowrap bgcolor='$color' align='right'>$updateCost$updateTime</td>
	    <td nowrap bgcolor='$color' align='right'>$footprint</td>
	    <td nowrap bgcolor='$color'>$command</td>
	  </tr>~;
}

# Generate rows for the Repositories
sub fragmentAdminRepositoryList {
	my @htmlrows = ();

	# Iterate over the Repositories
	foreach my $Repository ( $MetaData->getRepositories ) {
		push @htmlrows, fragmentAdminRepositoryRow( $Repository );
	}
	return '' unless @htmlrows;

	# Return the table
	my $joined = join "\n", @htmlrows;
	return qq~
		<table width="100%" border="0" cellspacing="0" cellpadding="0">
		  <tr>
		    <td bgcolor="#999999"> 
		      <table width="100%" border="0" cellspacing="1" cellpadding="2">
			 $joined
		      </table>
		    </td>
		  </tr>
		</table>
		~
}

# Generate the rows for a single Repository
sub fragmentAdminRepositoryRow {
	my @htmlrows = ();

	# Get and check the Repository
	my $Repository = isa( $_[0], 'CVSMonitor::MetaData::Repository' )
		? shift : return undef;

	# Get the label and CVSROOT
	my $label = CVSMonitor::HTML->escape( $Repository->getLabel );
	my $cvsroot = CVSMonitor::HTML->escape( $Repository->publicCvsroot );
	my $add_module_link = fragmentLink( 'Add Module', {
		cmd => 'viewAdminModuleNew',
		repository => $Repository->name,
		} );
	my $delete_link = fragmentLink( 'Delete Repository', {
		cmd => 'adminDeleteRepository',
		repository => $Repository->name,
		} );
	my $browse_link = fragmentLink( 'Browse', {
		cmd => 'viewBrowseRepository',
		repository => $Repository->name,
		} );
	my $update_link = fragmentLink( 'Update', {
		cmd => 'adminBackgroundUpdateRepository',
		repository => $Repository->name,
		} );

	# Add the row
	push @htmlrows, qq~
	  <tr> 
	    <td colspan='6' nowrap bgcolor='#FFFFFF'>
	      <b><font size='+1'>$label</b></font> ($cvsroot)<br>
	      $browse_link &middot; $update_link &middot; $add_module_link &middot; $delete_link
	    </td>
	  </tr>~;
	  
	# If the Repository has any modules, add the module headers
	if ( scalar $Repository->getModules ) {
		push @htmlrows, qq~
		  <tr class="table_header"> 
		    <td bgcolor="#CCCCCC" nowrap align="center"><b>Module</b></td>
		    <td bgcolor="#CCCCCC" nowrap align="center"><b>Status</b></td>
		    <td bgcolor="#CCCCCC" nowrap align="center"><b>Last Synchronised</b></td>
		    <td bgcolor="#CCCCCC" nowrap align="center"><b>Update Cost</b></td>
		    <td bgcolor="#CCCCCC" nowrap align="center"><b>Disk</b></td>
		    <td bgcolor="#CCCCCC" nowrap align="center"><b>Commands</b></td>
		  </tr>~;
	}

	# Iterate over the Modules, and add their rows
	foreach my $Module ( $Repository->getModules ) {
		push @htmlrows, fragmentAdminModuleRow( $Module );
	}

	@htmlrows;
}

sub fragmentAdminRefresh {
	# Get the command
	my $cmd = shift || $fin->{cmd};
	$cmd = 'viewAdminRepositoryList' unless $cmd =~ /^view/;

	# Return the image fragment
	return qq~<a href="$callback?cmd=$cmd"
	  ><img src="$static->{http}/static/images/refresh.gif" border="0"
	  ></a>~;
}

sub fragmentBrowseRepository {
	# Get and check the Repository
	my $Repository = shift or return undef;
	unless ( $Repository ) {
		Error( 'fragmentBrowseRepository was not passed a Repository' );
	}

	# Generate the rows for the Modules
	my @modulerows = ();
	foreach my $Module ( $Repository->getModules ) {
		# Get the bits and pieces
		my $display = CVSMonitor::HTML->escape( $Module->display );
		$display = $Module->available
			? fragmentLink( $display, {
				cmd    => 'viewBrowseModule',
				module => $Module->name,
				} )
			: $display;
		my $state = $Module->state;
		my $date = fragmentEventTime( $Module->dateRaw ) || '&nbsp';
		my $color = fragmentModuleColor( $Module );
		my $rss = fragmentLink( 'RSS', {
			cmd    => 'rssModule',
			module => $Module->name,
			} );

		push @modulerows, qq~
	          <tr bgcolor="$color"> 
	            <td nowrap>$display</td>
	            <td nowrap>$state</td>
	            <td nowrap>$date</td>
	            <td nowrap>$rss</td>
	          </tr>~;
	}

	# If there are no modules, add a note saying so
	unless ( @modulerows ) {
		push @modulerows, qq~
		  <tr bgcolor="#FFFFFF">
		    <td nowrap colspan='4'>No Modules Configured</td>
		  </tr>~;
	}

	# Get the label and CVSROOT and escape for display.
	# Only let them link to the Repository view if there are
	# more than 1 module in it, and all are available.
	my $modulecount = scalar $Repository->getModules;
	my $label = CVSMonitor::HTML->escape( $Repository->getLabel );
	$label = ($Repository->available and $modulecount >= 2) ?
		fragmentLink( $label, {
			cmd => 'viewBrowseRepository',
			repository => $Repository->name,
			} ) : $label;
	my $host = $Repository->host;
	my $modules = join "\n", @modulerows;

	# Return the completed repository fragment
	return qq~
	<table border="0" cellspacing="0" cellpadding="1">
	  <tr>
	    <td bgcolor="#999999"> 
	      <table border="0" cellspacing="1" cellpadding="3">
	        <tr bgcolor="#EEEEEE"> 
	          <td colspan="4" nowrap><font size="+1">$label</font>&nbsp;&nbsp;&nbsp;($host)</td>
	        </tr>
	        <tr bgcolor="#CCCCCC"> 
	          <td nowrap><b>Module Name</b></td>
	          <td nowrap><b>State</b></td>
	          <td nowrap><b>Last Updated</b></td>
	          <td nowrap>&nbsp;</td>
	        </tr>
	        $modules
	      </table>
	    </td>
	  </tr>
	</table>~;
}

# Generate a fragment for a set of ChangeSets.
# Map to single ChangeSets fragments, merge, and return.
sub fragmentChangeSets {
	my $ChangeSets = shift;
	join '<br>', map { fragmentChangeSet($_) } @$ChangeSets;
}

# Generate a fragment for a single ChangeSet
sub fragmentChangeSet {
	my $ChangeSet = isa( $_[0], 'CVSMonitor::MetaData::ChangeSet' )
		? shift : return undef;
	my $show_all_versions = shift;

	my $toomany = 0;
        my @Versions = $ChangeSet->getVersions;
	unless ( $show_all_versions ) {
		# Get the versions, and chop if there are too many.
		# Although 'too many' is considered over 30, we chop
		# back to 20, so we don't get a situation where there
		# are '1 version too many' messages.
		if ( scalar @Versions > 40 ) {
			$toomany = scalar @Versions - 20;
			@Versions = @Versions[0..19];
		}
	}

	# Iterate over the Versions and prepare the data for the table rows
	my @data = ();
	foreach my $Version ( @Versions ) {
		my $summary = $Version->summary;
		# Link it to view the Version...
		# but not if its a binary file, because at the moment
		# the only thing we show for a version is a diff.
		if ( $summary =~ /^([\d\.]+)/ and ! $Version->File->binary ) {
			my $versionlink = fragmentLink( $1, {
				cmd => 'viewBrowseVersion',
				module => $Version->Module->name,
				file => $Version->File->filename,
				version => $1,
				} );
			$summary =~ s/^([\d\.]+)/$versionlink/;
		}

		push @data, [ 
			fragmentLink( $Version->File->filename, {
				cmd    => 'viewBrowseFile', 
				module => $Version->Module->name,
				file   => $Version->File->filename,
				} ),
			$summary, 
			$Version->changeString 
			];
	}

	# Generate the row html
	my $joined_rows = CVSMonitor::HTML::Table->table_rows( \@data, {
		nowrap => 1,
		} );

	# Add the 'too many' line if needed
	if ( $toomany ) {
		my $show_entire_changeset = fragmentLink( 'Show Entire ChangeSet', {
			cmd => 'viewBrowseChangeSet',
			module => $ChangeSet->Module->name,
			id => $ChangeSet->id,
			} );
		$joined_rows .= qq~
		  <tr> 
		    <td colspan='3' bgcolor="#CCCCCC">Too many revisions to display, skipped $toomany. $show_entire_changeset</td>
		  </tr>~;
	}

	# Add the changeset id link
	my $changeset_link = fragmentLink( 'ChangeSet ' . $ChangeSet->id, {
		cmd => 'viewBrowseChangeSet',
		module => $ChangeSet->Module->name,
		id => $ChangeSet->id,
		} );

	# Get the author link
	my $author = fragmentEscapedLink( $ChangeSet->author, {
		cmd => 'viewBrowseAuthor',
		module => $ChangeSet->Module->name,
		author => $ChangeSet->author,
		} );

	# Fetch and linkify the comment
	my $comment = CVSMonitor::HTML->escape_multiline( $ChangeSet->comment );
	$comment = linkToTrackers( $ChangeSet->Module, $comment ) || $comment;

	# Add the remaining strings and generate the html
	my $date = fragmentEventTime( $ChangeSet->dateRaw, $ChangeSet->Module );
	my $totalchange = $ChangeSet->changeString;
	return qq~<table border="0" cellspacing="0" cellpadding="0">
	  <tr>
	    <td bgcolor="#999999"> 
	      <table border="0" cellspacing="1" cellpadding="3">
	        <tr class="table_header"> 
	          <td bgcolor="#CCCCCC">$changeset_link &middot; $date &middot; $author</td>
	          <td bgcolor="#CCCCCC" colspan="2"><b>$totalchange</b></td>
	        </tr>
	        <tr> 
	          <td bgcolor="#FFFFFF" colspan="3"><b>$comment</b></td>
	        </tr>
		    $joined_rows
	      </table>
	    </td>
	  </tr>
	</table>
	~;
}

# Create HTML for a graph
sub fragmentGraph {
	my $Source = shift or return undef;
	my $type = shift;
	my $grapher = shift;

	# Get the basic properties
	my $width = $grapher ? 800 : 400;
	my $height = $grapher ? 600 : 300;

	my $image = '';
	my $href = '';
	if ( $type ) {
		# We know the type, get the specific information
		$image = graphDynamicImageSrc( $type, $Source, $grapher );
		$href = graphGrapherLinkHref( $type, $Source );

	} else {
		# Get the graph options and display the first
		my $options = graphOptions( $Source, $grapher ) or return undef;
		$image = $options->[0]->{image};
		$href = $options->[0]->{href};
	}

	# Generate the image
	my $html = "<img name='graphic_image' src='$image' height='$height' width='$width' border='0'>";

	# Don't link if we are in the grapher already
	return $html if $grapher;

	# Wrap the image in the link to the grapher
	"<a name='graphic_link' href='$href'>$html</a>";
}

sub fragmentGraphSelector {
	my $Object = shift or return undef;
	my $grapher = shift;

	# Get the options
	my $graphs = graphOptions( $Object, $grapher ) or return undef;

	my @vars = ();
	my @options = ();
	my $i = 0;
	foreach ( @$graphs ) {
		push @vars, "Images[$i] = '$_->{image}';";
		push @vars, "HRefs[$i] = '$_->{href}';";
		push @options, [ $i, $_->{label} ];
		$i++;
	}

	# Create the drop box
	my $dropbox = CVSMonitor::HTML::Form->dropbox( 'graphic_selector', \@options, 0, {
		onChange => 'changeGraphic(this.value);',
		notnull => 1,
		} );

	my $varstring = join "\n", @vars;
	return qq~
		<script language='JavaScript'>
		<!--

		var Images = new Array();
		var HRefs = new Array();
		$varstring

		function changeGraphic(n) {
			document.graphic_image.src = Images[n];

			// Find the anchor
			for( i = 0; i < document.anchors.length; i++ ) {
				if ( document.anchors[i].name == 'graphic_link' ) {
					document.anchors[i].href = HRefs[n];
					return true;
				}
			}
			return true;
		}

		// -->
		</script>

		$dropbox
		~;
}

# Take a date, and make it linked
sub fragmentDateLink {
	my $date = shift;
	my $M = shift || $Module;
	return undef unless isa( $M, 'CVSMonitor::MetaData::Module' );

	my $dayregex = CVSMonitor::DateDepth->isolate_regex( 'day' );
	if ( $date =~ s/$dayregex// ) {
		$date = fragmentLink( $1, {
			cmd => 'viewBrowseDay',
			day => $1,
			module => $M->name,
			} ) . $date;
	}

	$date;
}

# Generate a fragment showing all the Versions for a file
sub fragmentFileHistory {
	my $File = isa( $_[0], 'CVSMonitor::MetaData::File' )
		? shift : return undef;

	# Get the Versions
	my $Versions = $File->getVersions or return undef;

	# Build the version lines
	my $html = '';
	my $rowcolor = '#FFFFFF';
	foreach my $Version ( @$Versions ) {
		my $date = fragmentDateLink( $Version->date, $File->Module );
		my $author = fragmentEscapedLink( $Version->author, {
			cmd => 'viewBrowseAuthor',
			module => $Version->Module->name,
			author => $Version->author,
			} );
		my $summary = $Version->summary;
		if ( $summary =~ /^([\d\.]+)/ ) {
			# Link it to view the Version
			my $versionlink = fragmentLink( $1, {
				cmd => 'viewBrowseVersion',
				module => $mod,
				file => $File->filename,
				version => $Version->revision,
				} );
			$summary =~ s/^([\d\.]+)/$versionlink/;
		}
		my $change = $Version->changeString;
		my $comment = CVSMonitor::HTML->escape_multiline( $Version->comment );
		$comment = linkToTrackers( $Version->Module, $comment ) || $comment;

		$rowcolor = ($rowcolor eq '#FFFFFF') ? '#F4F4F4' : '#FFFFFF'	;
		$html .= qq~<tr> 
			  <td bgcolor="$rowcolor" nowrap>$date</td>
		          <td bgcolor="$rowcolor" nowrap>$author</td>
		          <td bgcolor="$rowcolor" nowrap>$summary</td>
		          <td bgcolor="$rowcolor" nowrap>$change</td>
		          <td bgcolor="$rowcolor">$comment</td>
			</tr>
		        ~;
	}

	# Wrap in the main html
	my $file = $File->filename;
	return qq~
		<table border="0" cellspacing="0" cellpadding="1">
		  <tr>
		    <td bgcolor="#999999">
		      <table border="0" cellspacing="1" cellpadding="3">
		        <tr bgcolor="#CCCCCC"> 
		          <td colspan="5"><b>$file</b></td>
		        </tr>
			$html
		      </table>
		    </td>
		  </tr>
		</table>
		~;
}

# Create a fragment containing a list of selectable files
sub fragmentFiles {
	my $Files = shift or return '';
	my $opt = isa( $_[0], 'HASH' ) ? shift : { module => 1 };

	# Populate the table
	my @data = ( [ ($opt->{module} ? 'Module' : ()), 'File Name', 'Created', 'Revisions', 'Last Change' ] );
	foreach my $File ( @$Files ) {
		# Link to the Module
		my $module = $opt->{module}
			? fragmentLink( $File->Module->name, {
			cmd => 'viewBrowseModule',
			module => $File->Module->name,
			} ) : '';

		# Create the file link
		my $filename = fragmentLink( $File->filename, {
			cmd => 'viewBrowseFile',
			module => $File->Module->name,
			file => $File->filename,
			} );

		# Add the creation date
		my $created = fragmentDateLink( $File->getFirstVersion->date, $File->Module );

		# Get the version count
		my $versions = "<div align='right'>" . $File->versions . '</div>';

		# When was the last change
		my $ChangeSet = $File->getRecentVersion->ChangeSet;
		my $change = $ChangeSet
			? ( fragmentDateLink( $ChangeSet->date, $ChangeSet->Module )
				. ' &middot; '
				. fragmentLink( 'ChangeSet ' . $ChangeSet->id, {
					cmd => 'viewBrowseChangeSet',
					module => $File->Module->name,
					id => $ChangeSet->id,
					} ) )
			: 'Unknown';

		# Add to the row
		push @data, [ ($opt->{module} ? $module : ()), $filename, $created, $versions, $change ];
	}

	CVSMonitor::HTML::Table->table( \@data, {
		top         => 1,
		topclass    => 'table_header',
		} );
}


# Generate a fragment for the totals of the Authors
sub fragmentAuthorTotals {
	my $AuthorTotals = isa( $_[0], 'CVSMonitor::AuthorTotals' ) ? shift : return undef;

	# Get and check the options
	my $options = shift || {};
	unless ( defined $options->{sortBy} ) {
		$options->{sortBy} = $fin->{author_sort_by} || 'karma';
	}

	# Get the authors sort by a AuthorTotal legal value
	my $Authors = $AuthorTotals->getAuthorsSorted( $options->{sortBy} ) or return undef;
	@$Authors = grep { $_->{changes} } @$Authors;
	return undef unless scalar @$Authors;
	unless ( $options->{sortBy} eq 'author' ) {
		@$Authors = reverse @$Authors;
	}

	# Format into html rows
	my @data = ( [ 
		fragmentLinkBack( 'Author',  { author_sort_by => 'author'  } ),
		fragmentLinkBack( 'Karma',   { author_sort_by => 'karma'   } ),
		fragmentLinkBack( 'Commits', { author_sort_by => 'commits' } ),
		fragmentLinkBack( 'Added',   { author_sort_by => 'up'      } ),
		fragmentLinkBack( 'Removed', { author_sort_by => 'down'    } ),
		fragmentLinkBack( 'Change',  { author_sort_by => 'change'  } ),
		] );
	my $count = 0;
	foreach my $Author ( @$Authors ) {
		# Apply the limit checking if applicable
		if ( $options->{limit} ) {
			last if ++$count > $options->{limit};
		}

		# Add the html
		my $name = $fin->{module}
			? fragmentEscapedLink( $Author->{author}, {
				cmd    => 'viewBrowseAuthor',
				module => $Module->name,
				author => $Author->{author},
				} )
			: CVSMonitor::HTML->escape($Author->{author});
		push @data, [ $name, 
			commify( $Author->{karma} ),
			commify( $Author->{commits} ),
			commify( $Author->{up} ),
			commify( $Author->{down} ),
			commify( $Author->{change} )
			];
	}

	CVSMonitor::HTML::Table->table( \@data, {
		normalalign => 'right',
		top         => 1,
		topclass    => 'table_header',
		outerwidth  => 0,
		left        => 1,
		boldleft    => 0,
		leftcolor   => '#EEEEEE',
		leftalign   => 'left',
		nowrap      => 1,
		} );
}

# From an Activity object generate a fragment 
# for the most recent ChangeSets.
sub fragmentActivityRecent {
	my $Activity = isa( $_[0], 'CVSMonitor::MetaData::Activity' ) ? shift : return undef;

	# Get the most most recent ChangeSets
	my $ChangeSets = $Activity->getRecentChangeSets( $Config->getGuiChangesets );
	return undef unless defined $ChangeSets;
	return 'No ChangeSets' unless $ChangeSets;

	fragmentChangeSets( $ChangeSets );
}

# From an Activity object generate a fragment
# for the most important ChangeSets
sub fragmentActivityImportant {
	my $Activity = isa( $_[0], 'CVSMonitor::MetaData::Activity' ) ? shift : return undef;

	# Get the most important ChangeSets
	my $ChangeSets = $Activity->getImportantChangeSets( $Config->getGuiChangesets );
	return undef unless defined $ChangeSets;
	return 'No ChangeSets' unless $ChangeSets;

	fragmentChangeSets( $ChangeSets );
}

# Turn all the params into a collection of hidden fields
sub fragmentHiddenFields {
	my $query = shift || CGI->new();
	return undef unless isa( $query, 'CGI' );

	# Get the params and map them to HTML
	my $html = '';
	foreach my $param ( $query->param ) {
		next if $param eq '_admin_pass';
		next if $param eq '_submit';
		my @vals = $query->param( $param );
		$html .= join "\n", 
			map { "<input type='hidden' name='$param' value='$_'>" }
			@vals;
	}

	$html;
}

# Decides the availability color flag for a module
sub fragmentModuleColor {
	my $Module = shift;
	return '#FF9999' unless isa( $Module, 'CVSMonitor::MetaData::Module' );
	my $state = $Module->state;

	# Return the appropriate color
	if ( $Module->broken ) {
		return '#FF9999'; # Red
	} elsif ( $state eq 'Unavailable' or $state eq 'Checking Out' ) {
		return '#FFCC00'; # Orange
	} elsif ( $state eq 'Updating' or $state eq 'Available' ) {
		return '#99FF99'; # Green
	} else {
		return '#FF9999'; # Red
	}
}





#####################################################################
# Toolbar fragment and related function

# Add a toolbar by name
sub addToolbar {
	my $name = lc shift;

	# Switch on tolbar name
	if ( $name eq 'main' ) {
		$gui->{toolbars}->{main} = fragmentToolbarMain( @_ );
	} elsif ( $name eq 'search' ) {
		$gui->{toolbars}->{search} = fragmentToolbarSearch( @_ );
	} elsif ( $name eq 'repository' ) {
		$gui->{toolbars}->{repository} = fragmentToolbarRepository( @_ );
	} elsif ( $name eq 'module' ) {
		$gui->{toolbars}->{module} = fragmentToolbarModule( @_ );
	} elsif ( $name eq 'file' ) {
		$gui->{toolbars}->{file} = fragmentToolbarFile( @_ );
	} elsif ( $name eq 'version' ) {
		$gui->{toolbars}->{version} = fragmentToolbarVersion( @_ );
	} else {
		Error( "Unknown toolbar '$name'" );
	}

	1;
}

# Clear existing toolbars
sub clearToolbars {
	$gui->{toolbars} = {};
	1;
}

# Go through the toolbars in the correct order and merge them
sub fragmentToolbars {
	my $bar = fragmentBlackLine();
	my $toolbar = join $bar, 
		map { $gui->{toolbars}->{$_} }
		grep { defined $gui->{toolbars}->{$_} }
		qw{main search repository module file version};
	"$bar$toolbar$bar";
}

# Create the main menu
sub fragmentToolbarMain {
	my $disabled = shift;

	# Are there any repositories
	my $repositories = ( $MetaData and scalar $MetaData->getRepositories ) ? 1 : 0;
	my $active = ( $MetaData and scalar grep {  $_->available } $MetaData->getModules ) ? 1 : 0;

	# Create the links
	my $browse = $disabled ? 'Browse' : $repositories
		? fragmentLink( 'Browse', { cmd => 'viewBrowse' } )
		: fragmentLinkDisabled( 'Browse' );
	my $search = $disabled ? 'Search' : $active
		? fragmentLink( 'Search', { cmd => 'viewSearch' } )
		: fragmentLinkDisabled( 'Search' );
	my $statistics = $disabled ? 'Statistics' : 0 # $repositories
		? fragmentLink( 'Statistics', { cmd => 'viewStatistics' } )
		: fragmentLinkDisabled( 'Statistics' );
	my $about = $disabled ? 'About'
		: fragmentLink( 'About', { cmd => 'viewAbout' } );
	my $personal = $disabled ? 'User Settings' : 1
			? fragmentLinkDisabled( 'User Settings' )
			: fragmentLink( 'User Settings', { cmd => 'viewPersonalSettings' } );
	my $admin = ( $disabled or ! defined $Config ) ? 'Admin'
		: fragmentLink( 'Admin', { cmd => 'viewAdmin' } );
	my $help = "<a href='http://ali.as/devel/cvsmonitor/help/release-0-6.html' target='_help'>Help</a>";

	# Create and return the html
	my $toolbar_class = $disabled
		? 'top_toolbar_disabled'
		: 'top_toolbar';
	return qq~
	  <table width="100%" border="0" cellspacing="0" cellpadding="2">
        <tr>
          <td class="top_toolbar_title" bgcolor="#CCCCCC" width="67" nowrap>&nbsp;Main</td>
          <td class="$toolbar_class" bgcolor="#CCCCCC" nowrap>&nbsp;$browse &middot; $search <!-- Don't show proposed features &middot; $statistics --></td>
          <td class="$toolbar_class" bgcolor="#CCCCCC" align="right" nowrap><!-- $personal &middot; --> $admin &middot; $help &middot; $about&nbsp;&nbsp;</td>
        </tr>
      </table>
      ~;
}

# Create the module menu
sub fragmentToolbarModule {
	return undef unless $Module;
	my $Rep = $Module->Repository or return undef;

	# Generate values as needed
	my $date = fragmentEventTime( $Module->dateRaw, $Module );

	# Link to the repository if there are more than one module
	my $repository = $Rep->available
		? fragmentLink( $Rep->getLabel, {
			cmd => 'viewBrowseRepository',
			repository => $Rep->name,
			} )
		: $Rep->getLabel;

	# Link to the Module summary
	my $module = fragmentLink( $Module->shortName, {
		cmd => 'viewBrowseModule',
		module => $Module->name,
		} );

	# Link to search for a file and a changeset in this module
	my $file_search = fragmentLink( 'File Search', {
		cmd => 'viewSearchFile',
		space => $Module->name,
		} );
	my $changeset_search = fragmentLink( 'ChangeSet Search', {
		cmd => 'viewSearchChangeSet',
		space => $Module->name,
		} );

	# Return the html
	return qq~
	  <table width="100%" border="0" cellspacing="0" cellpadding="2">
        <tr> 
          <td class="top_toolbar_title" bgcolor="#CCCCCC" width="67" nowrap>&nbsp;Module</td>
          <td class="top_toolbar" bgcolor="#CCCCCC" nowrap>&nbsp;$repository &middot; $module</td>
          <td class="top_toolbar" bgcolor="#CCCCCC" nowrap align="center">$file_search &middot; $changeset_search</td>
          <td class="top_toolbar" align="right" bgcolor="#CCCCCC" nowrap><span class='text_subtle'>Updated: $date</span>&nbsp;&nbsp;</td>
        </tr>
      </table>
      ~;
}

# Create the repository menu, which is basically links to individual modules
sub fragmentToolbarRepository {
	return undef unless $Repository;

	# Get the strings we will need
	my $repository_label = $Repository->getLabel;

	# Generate a joined string of links to the modules
	my $module_links = join ' &middot; ', map { 
		$_->available
			? fragmentLink( $_->display, {
				cmd => 'viewBrowseModule',
				module => $_->name,
				} )
			: fragmentDisabled( $_->display )
		} $Repository->getModules;

	return qq~
	  <table width="100%" border="0" cellspacing="0" cellpadding="2">
        <tr> 
          <td class="top_toolbar_title" bgcolor="#CCCCCC" width="67" nowrap>&nbsp;Module</td>
          <td class="top_toolbar" bgcolor="#CCCCCC" nowrap>&nbsp;$repository_label</td>
          <td class="top_toolbar" align="right" bgcolor="#CCCCCC" nowrap>$module_links&nbsp;&nbsp;</td>
        </tr>
      </table>
	~;
}

# Create the search menu
sub fragmentToolbarSearch {

	# Create the links
	my $changeset = fragmentLink( 'ChangeSet', { cmd => 'viewSearchChangeSet' } );
	my $files = fragmentLink( 'File', { cmd => 'viewSearchFile' } );

	# Return the fragment
	return qq~<tr> 
	  <table width="100%" border="0" cellspacing="0" cellpadding="2">
        <tr> 
          <td class="top_toolbar_title" bgcolor="#CCCCCC" width="67" nowrap>&nbsp;Search</td>
          <td class="top_toolbar" bgcolor="#CCCCCC" nowrap>&nbsp;Search for: $changeset &middot; $files</td>
        </tr>
      </table>
	~;
}

# Generate the toolbar fragment for the file section
sub fragmentToolbarFile {
	my $File = isa( $_[0], 'CVSMonitor::MetaData::File' )
		? shift : return undef;

	# Create the links
	my $filename = $File->filename;
	my $history = fragmentLink( 'History', {
		cmd => 'viewBrowseFileHistory',
		module => $File->Module->name,
		file => $filename,
		} );
	my $annotation = $File->binary
		? fragmentLinkDisabled( 'Annotation' )
		: fragmentLink( 'Annotation', {
			cmd => 'viewBrowseFileAnnotation',
			module => $File->Module->name,
			file => $filename,
			} );
	my $blame = $File->binary
		? fragmentLinkDisabled( 'Blame' )
		: fragmentLink( 'Blame', {
			cmd => 'viewBrowseFileBlame',
			module => $File->Module->name,
			file => $filename,
			} );

	# Generate and return
	return qq~
	  <table width="100%" border="0" cellspacing="0" cellpadding="2">
	    <tr> 
	      <td class="top_toolbar_title" bgcolor="#CCCCCC" width="67" nowrap>&nbsp;File</td>
	      <td class="top_toolbar" bgcolor="#CCCCCC" nowrap>&nbsp;$history &middot; $blame &middot; $annotation</td>
	      <td class="top_toolbar" bgcolor="#CCCCCC" align="right" nowrap><span class='text_subtle'>$filename</span>&nbsp;&nbsp;</td>
	    </tr>
	  </table>
	~;
}

# Create a toolbar for a revision
sub fragmentToolbarVersion {
	my $Version = isa( $_[0], 'CVSMonitor::MetaData::Version' )
		? shift : return undef;

	# Generate the links for the toolbar
	# Add the previous and next revision links
	my $Previous = $Version->previous;
	my $previous_link = $Previous
		? fragmentLinkBack( $Previous->revision, {
			version => $Previous->revision 
			} )
		: 'None';
	my @Next = $Version->next();
	my $next_links = @Next
		? join( ', ', map { 
			fragmentLinkBack( $_->revision, {
				version => $_->revision
				} )
			} @Next )
		: 'None';

	# Add the Diff link
	my $diff = fragmentLink( 'Diff', {
		cmd => 'viewBrowseVersionDiff',
		module => $Version->Module->name,
		file => $Version->File->filename,
		version => $Version->revision,
		} );

	# Generate and return
	my $revision = $Version->revision;
	return qq~
	  <table width="100%" border="0" cellspacing="0" cellpadding="2">
	    <tr> 
	      <td class="top_toolbar_title" bgcolor="#CCCCCC" width="67" nowrap>&nbsp;Revision</td>
	      <td class="top_toolbar" bgcolor="#CCCCCC" nowrap>&nbsp;$diff</td>
	      <td class="top_toolbar" bgcolor="#CCCCCC" align="right" nowrap>$previous_link <-- $revision --> $next_links&nbsp;&nbsp;</td>
	    </tr>
	  </table>
	~;
}

# HTML fragment for configurable time format, absolute GMT, or relative
sub fragmentEventTime {
	my $epoch = shift or return '';

	# Handle a 'mixed' setting
	my $timeformat = $Config->getGuiTimeformat;
	if ( $timeformat eq 'smart' ) {
		$timeformat = (time - $epoch) < (31 * 24 * 60 * 60) ? 'relative' : 'GMT';
	}

	if ( $timeformat eq 'GMT' ) {
		# Create an ISO formatted GMT date
		my @t = gmtime $epoch;
		$t[4] += 1;
		$t[5] += 1900;
		@t[0..4] = map { sprintf '%02d', $_ } @t[0..4];
		my $string = "$t[5]/$t[4]/$t[3] $t[2]:$t[1]:$t[0] GMT";

		# Link the string if needed
		if ( isa( $_[0], 'CVSMonitor::MetaData::Module' ) ) {
			$string = fragmentDateLink( $string, shift );
		}

		return $string;

	} elsif ( $timeformat eq 'relative' ) {
		# Use Time::Duration to do the relative times
		return Time::Duration::ago( time - $epoch );

	} else {
		return 'Unknown time format';
	}
}





#####################################################################
# Other interace HTML related functions

sub addAdminMenu {
	my $parser = shift or return undef;

	# If we have a MetaData object, show the Repository menu header and
	# the Add Repository options.
	if ( $MetaData ) {
		$parser->{menu_repository_list} = '- '
			. fragmentLink( 'Repository List', {
				cmd => 'viewAdminRepositoryList',
				} );
		$parser->{menu_add_repository} = '- ' 
			. fragmentLink( 'Add Repository', {
				cmd => 'viewAdminRepositoryNew',
				} );
	} else {
		$parser->{menu_repository_list} = fragmentLinkDisabled( '- ' . 'Repository List' );
		$parser->{menu_add_repository} = fragmentLinkDisabled( '- ' . 'Add Repository' );
	}

	# If there are one or more repositories, show the Add Module option 
	# and the refresh view option.
	if ( $MetaData and scalar $MetaData->getRepositories ) {
		$parser->{menu_add_module} = '- ' 
			. fragmentLink( 'Add Module', {
				cmd => 'viewAdminModuleNew',
				} );
	} else {
		$parser->{menu_add_module} = fragmentLinkDisabled( '- ' . 'Add Module' );
	}

	# If there are any Available, non-broken modules, show the
	# update all modules menu item
	my @Modules = $MetaData ? $MetaData->getModules : ();
	if ( scalar grep { ! $_->broken } grep { $_->state eq 'Available' } @Modules ) {
		$parser->{menu_update_all} = '- ' . fragmentLink( 'Update All', {
			cmd => 'adminBackgroundUpdateAll',
			} );
	} else {
		$parser->{menu_update_all} = fragmentLinkDisabled( '- ' . 'Update All' );
	}

	# If there are no locked, unbroken modules, show the 
	# Reset MetaData Cache
	# if ( scalar @Modules == 0 or scalar grep { $_->locked and $_->lockHolderExists } @Modules ) {
	#	# Show a greyed out link
	#	$parser->{menu_reset_cache} = fragmentLinkDisabled( '- ' . 'Reset MetaData Cache' );
	#} else {
	#	# Show the Reset MetaData Cache link
	#	$parser->{menu_reset_cache} = '- '
	#		. fragmentLink( 'Reset MetaData Cache', {
	#			cmd => 'adminResetMetaDataCache',
	#			} ) ;
	#}

	1;
}

sub fragmentBlackLine {
	my $static = $static->{http} || '';
	return qq~
	<table width="100%" border="0" cellspacing="0" cellpadding="0">
	  <tr>
	    <td bgcolor="#000000"><img src="$static/static/images/pixel.gif" width="1" height="1"></td>
	  </tr>
	</table>
	~;
}





#####################################################################
# Utility Functions

# Find a search space as an Activity object
sub findSearchSpace {
	# First, get all the 'space' values
	my $query = CGI->new();
	my @spaces = $query->param( 'space' );
	return 0 unless scalar @spaces;

	# Next, map them to a list of Module objects.
	my %Modules = ();
	foreach my $space ( @spaces ) {
		if ( $space =~ /\./ ) {
			# It should be a module name
			my $Module = $MetaData->getModule( $space );
			unless ( $Module ) {
				Error( "No such module '$space' while building search space" );
			}
			next if $Modules{$Module->name};
			$Modules{$Module->name} = $Module;
		} else {
			# It should be a repository
			my $Repository = $MetaData->getRepository( $space );
			unless ( $Repository ) {
				Error( "No such repository '$space' while building search space" );
			}

			# Add all it's modules
			foreach my $Module ( $Repository->getModules ) {
				next if $Modules{$Module->name};
				$Modules{$Module->name} = $Module;
			}
		}
	}

	# Do we have any modules?
	return 0 unless scalar keys %Modules;

	# Merge the Modules into a single Activity object
	my $Activity = CVSMonitor::MetaData::Activity->new();
	Error( 'Failed to create empty Activity object while building search space' ) unless $Activity;
	foreach ( keys %Modules ) {
		$Activity->add( $Modules{$_} )
			or Error( "Failed to add module '$_' while building search space" );
	}

	# Return the Activity for the search space
	$Activity;
}

# We expect a module to exist.
sub checkModule {
	my $dieOnError = (defined $_[0] and $_[0] eq 'die') ? 1 : '';

	# Get the module
	unless ( $fin->{module} ) {
		$dieOnError 
			? die( "No module name was specified" )
			: Error( "No module name was specified" );
	}
	$mod = $fin->{module};
	$Module = $MetaData->getModule($mod);
	unless ( $Module ) {
		$dieOnError
			? die( "The module '$mod' does not exist" )
			: Error( "The module '$mod' does not exist" );
	}
}

# We except a module to exist and be available.
sub checkAvailableModule {
	my $type = shift;
	my $dieOnError = (defined $type and $type eq 'die') ? $type : '';
	checkModule( $dieOnError );

	# Has the module been checked out
	unless ( $Module->available('read') ) {
		$dieOnError
			? die( "The module '$mod' has not been checkout out" )
			: Error( "The module '$mod' has not been checkout out" );
	}

	# Module related, add the module toolbar
	addToolbar( 'module' );
}

# As above, but for Repositories
sub checkRepository {
	my $dieOnError = ( defined $_[0] and $_[0] eq 'die' ) ? shift : '';

	# Get the module
	unless ( $fin->{repository} ) {
		$dieOnError 
			? die( "No repository name was specified" )
			: Error( "No repository name was specified" );
	}
	$rep = $fin->{repository};
	$Repository = $MetaData->getRepository($rep);
	unless ( $Repository ) {
		$dieOnError
			? die( "The repository '$rep' does not exist" )
			: Error( "The repository '$rep' does not exist" );
	}
}

# We except a module to exist and be available.
sub checkAvailableRepository {
	my $type = shift;
	my $dieOnError = (defined $type and $type eq 'die') ? $type : '';
	checkRepository( $dieOnError );

	# Are all of the modules in it available
	foreach my $Module ( $Repository->getModules ) {
		unless ( $Module->available('read') ) {
			$dieOnError
				? die( "One or more modules in the repository '$rep' are not available" )
				: Error( "One or more modules in the repository '$rep' are not available" );
		}
	}

	# Add the toolbar for it
	addToolbar( 'repository' );
}

# Check that we have access to the MetaData repository
sub checkMetaData {
		# Should we hold keep an exclusive ( write ) lock on the
		# MetaData file. We only need this if we are going to do
		# something admin'y.
		my $lockIndex = $cmd =~ /^admin/ ? 'lock' : 'readonly';

		# Check if the index exists
		my $root = $Config->getHostScratch;
		my $base = CVSMonitor::MetaData->_checkWorkArea( $root ) 
			or Error( "Error checking scratch area" );
		my $file = File::Spec->catfile( $base, 'index.dat' ) 
			or Error( "Error determining index file name" );
		if ( -e	$file ) {
			# Try to load it
			unless ( $MetaData = CVSMonitor::MetaData->load( $root, $lockIndex ) ) {
				$message = CVSMonitor::MetaData->errstr;
				ErrorBrokenCache();
			}
		} else {
			# Create a new cache
			unless ( $MetaData = CVSMonitor::MetaData->new( $root, $lockIndex ) ) {
				Error( "Failed to create new MetaData cache", 
					CVSMonitor::MetaData->errstr );
			}
		}
}

# Put commas into numbers
sub commify {
	my $comma = shift;
	1 while $comma =~ s/^([-+]?\d+)(\d{3})/$1,$2/;
	$comma;
}

# Add line numbers to text ( assume local newlines )
# Do it by reference to reduce memory overhead.
sub linenumberify {
	my $text = defined $_[0] ? shift : return undef;
	return undef if ref $text;
	return undef unless length $text;

	# First, determine the number of lines and the format
	# for the sprintf.
	my $lines = scalar( my @newlines = $text =~ /\n/g ) + 1;
	my $sformat = '%' . length($lines) . 'd: ';

	# Insert all the line labels
	my $line = 0;
	$text =~ s/(^|\n)/$1 . sprintf( $sformat, ++$line )/ge;

	$text;
}

# Take a fragment of text, from a module, and add the links to external systems
sub linkToTrackers {
	my $Module = isa( $Module, 'CVSMonitor::MetaData::Module' )
		? shift : return undef;
	my $text = shift;

	# Get the tracker information
	my $trackers = $Module->trackers;
	unless ( $trackers ) {
		# No trackers, nothing to do
		return $text;
	}

	# Let each tracker go over the text and modify it
	foreach my $keyword ( keys %$trackers ) {
		my $url = $trackers->{$keyword};

		# Create the search regex
		my $search = qr/\b($keyword(?:\:?\s*|\s+)\#?(\d+))/i;

		# Do the search and replace
		$text =~ s/$search/&_linkToTrackersReplacer( $url, $1, $2 )/eg;
	}

	$text;
}

sub _linkToTrackersReplacer {
	my $url = shift;
	my $text = shift;
	my $id = shift;
	$url =~ s/\%n/$id/g;

	# Create the link
	my $target = '_' . AppLib::String->random;
	"<a href='$url' target='$target'>$text</a>";
}






#####################################################################
# Page Library functions and Error Handling

sub new_parser {
	my $parser = AppLib::Parser->new();

	# Add all the default variables
	$parser->{charset} = $CVSMonitor::CHARSET;
	$parser->{version} = $versionstring;
	$parser->{callback} = $callback;
	$parser->{module} = $mod;
	$parser->{repository} = $rep;
	$parser->{admin_title_right} = '&nbsp;';

	# Add the appropriate menu
	if ( $cmd =~ /^(?:admin|viewAdmin)/ ) {
		# Do Nothing 
	} elsif ( $Module ) {
		# Show the module menu
		$parser->{toolbar} = fragmentToolbarModule();
	}

	# Add some stuff any time a module or repository is mentioned
	if ( $Module ) {
		$parser->{module_display} = $Module->display;
		$parser->{module_date} = fragmentEventTime( $Module->dateRaw );
	}
	if ( $Repository ) {
		$parser->{repository_label} = $Repository->getLabel;
	}

	# Add the global message
	$parser->{message} = $message if $message;

	$parser;
}

# Create and return a page object, checking for errors
sub page {
	my $page = shift;
	my $parser = shift;
	$parser = new_parser() unless $parser;

	if ( $message ) {
		$parser->{message} = $message;
	}

	# Add the toolbar if needed
	$parser->{toolbar} = $parser->{toolbars} = fragmentToolbars();

	# Add the title of the page
	$parser->{title} = $gui->{title}
		? "CVS Monitor - $gui->{title}"
		: "CVS Monitor";

	# Create the page
	my $Object = AppLib::Page->new( $page, $parser );
	unless ( defined $Object ) {
		Error( "Error trying to show page<br>" . AppLib::Error->errstrHTML );
	}
	$Object;
}

# Handle errors
sub Error {
	# If we are in image mode, just return the error image
	if ( $imagemode ) {
		ErrorImage();
	}

	# Get the correct message
	my $message = join "<br>", @_;
	$message =~ s/\n/<br>/g;
	$message = 'Unknown error' unless $message;

	# Create the parser
	my $parser = new_parser();
	$parser->{message} = $message;

	# Try to show the error page
	my $Page = page( 'Error/Generic', $parser );
	if ( $Page ) {
		# Show the page
		$Page->display;
	} else {
		# First error call didn't work, let's do one that is 
		# completely self contained.
		$message .= '<br><br>In addition, an error occurred while'
			. ' trying to display the error page<br>'
			. AppLib::Page->errstr;
		print "Content-type: text/html\n\n";
		print "<HTML><HEAD><TITLE>Error</TITLE></HEAD>\n";
		print "<BODY><H1>Error</H1>\n";
		print "<i>$message</i>\n";
		print "</BODY></HTML>\n";
	}
	exit();
}

sub ErrorBrokenCache {
	my $parser = new_parser();
	my $Page = page( 'Error/BrokenCache', $parser );
	$Page->display;
	exit();
}

# Provide an error that will work in the context
# of when we are being called as an image
sub ErrorImage {
	# Simply redirect to the static error image
	CGI::redirect( "$static->{http}/static/images/error.gif" );
	exit();
}
