package CVSMonitor::Security;

# The CVSMonitor::Security package implements the ( this ) security
# layer for CVS Monitor.
#
# While this module provides a password handling, checking and 
# caching facility, it does not enfore password integrity.

use strict;
use CGI ();





#####################################################################
# Global settings

use vars qw{$DISABLE $PASSWORD};
BEGIN {
	# Change the value of $DISABLE to true to disable security
	$DISABLE = 0;

	# Set the default password here
	$PASSWORD = '';
}

# Change the password
sub setPassword {
	my $class = shift;
	$PASSWORD = shift;
	1;
}





#####################################################################
# Main authentication methods

# Checks to see if a password matches, or checks for the cookie.
# Returns 1 if authenticated.
# Returns 0 if no password given.
# Returns undef if password is incorrect.
sub authenticate {
	my $class = shift;
	return 1 if $DISABLE;
	return 1 if $PASSWORD eq '';

	# Have they passed a password directly
	my $pass = '';
	if ( @_ ) {
		$pass = shift;
	} else {
		# Try for the cookie
		$pass = CGI->new()->cookie( '-name' => 'CVSMonitor' );
	}

	# Do they have a password
	if ( $pass ) {
		# Does it match
		return $pass eq $PASSWORD ? 1 : undef;
	} else {
		# No password
		return 0;
	}
}

# Save a password to their browser.
# We save the password by adding it as a cookie
# to the next page that gets returned to the user.
sub savePassword {
	my $class = shift;
	my $pass = shift;

	# Create the cookie for the password
	my $cookie = CGI->new()->cookie(
		'-name' => 'CVSMonitor',
		'-value' => $pass,
		'-expires' => '+2h',
		'-path' => '/',
		);

	# Add it to the next page
	AppLib::PageFactory->setCookie( $cookie );
}

1;

__END__

# Copyright (C) 2002-2004 Adam Kennedy
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Should you wish to utilise this software under a different licence, 
# please contact the author.
